# -*- coding: utf-8 -*-

"""Acclerated optimization using MOPAC for Hessians."""

import json
import logging
from pathlib import Path
import pkg_resources
import sys
import textwrap
import traceback

import numpy as np

# import optking
from scipy.optimize import minimize, OptimizeResult
from scipy.optimize._optimize import (
    _prepare_scalar_function,
    _status_message,
    _epsilon,
    vecnorm,
    _line_search_wolfe12,
    _LineSearchError,
)

from .optimizer import Optimizer
import psi4_step
import seamm
import seamm.data
from seamm_util import units_class, Q_
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("psi4")

optking_logger = logging.getLogger("optking")
optking_logger.setLevel(logging.WARNING)


def norm(V):
    return np.linalg.norm(V)


def abs_max(V):
    return max(abs(elem) for elem in V)


def abs_min(V):
    return min(abs(elem) for elem in V)


def rms(V):
    return np.sqrt(np.mean(V**2))


class AcceleratedOptimization(psi4_step.Energy):
    def __init__(
        self,
        flowchart=None,
        title="Accelerated Optimization",
        extension=None,
    ):
        """Initialize the node"""

        logger.debug("Creating AcceleratedOptimization {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self._calculation = "optimization"
        self._model = None
        self._metadata = psi4_step.metadata
        self.parameters = psi4_step.AcceleratedOptimizationParameters()

        self.description = "A geometry optimization"
        self._E = None
        self._derivatives = None
        self._triangle = None  # Triangule Hessian matrix
        self.prolog = None
        self.epilog = None
        self.iteration = 0
        self._restart_wfn = None

        self._counters = {}

    def description_text(self, P=None):
        """Prepare information about what this node will do"""

        if not P:
            P = self.parameters.values_to_dict()

        text = super().description_text(P=P, calculation_type="Geometry optimization")

        added = "The geometry optimization will use the {optimization method} "
        if P["max geometry steps"] == "default":
            added += "method, using the default maximum number of steps, which"
            added += " is based on the system size."
        else:
            added += "method, with no more than {max geometry steps} steps."

        if P["geometry convergence"] == "Custom":
            added += " The convergence criterion is"
        else:
            added += " The convergence criterion is '{geometry convergence}'."

        if P["recalc hessian"] != "never":
            added += (
                " The Hessian will be recalculated with MOPAC every {recalc hessian}"
                " steps. Note that calculating the second derivatives with MOPAC is "
                "not expensive!"
            )

        return text + "\n" + __(added, **P, indent=4 * " ").__str__()

    def input_text(self, node0, memory, restart=None):
        """Create the first and last parts of the Psi4 input.

        Parameters
        ----------
        node0 : seamm.Node
            The first node in the Psi4 subflowchart

        memory : str
            Amount of memory to allocate Psi4

        n_threads : int
            Number of threads to use in Psi4

        restart : str = None
            A wavefunction file to restart from
        """
        prolog = textwrap.dedent(
            f"""
            import json
            import numpy as np
            import pprint

            memory {memory}

            """
        )

        node = node0
        tmp = []
        while node is not None:
            if node == self:
                break
            text = node.get_input()
            tmp.append(text)

            tmp.append("clean()")
            tmp.append("clean_variables()")
            # tmp.append('clean_options()')

            node = node.next()

        # Add our input
        tmp.append("")
        tmp.append("#" * 80)
        tmp.append(f"# {self.header}")
        tmp.append("#" * 80)
        tmp.append("")

        # Add in the input from the energy part of things
        tmp.append(super().get_input(calculation_type="gradient", restart=restart))

        # Write out the wavefunction
        tmp.append("")
        tmp.append("wfn.to_file('wfn')")

        # Write out the final structure
        tmp.append("")
        tmp.append("# Write the final structure to disk")
        tmp.append("molecule = get_active_molecule()")
        tmp.append("tmp = molecule.to_dict()")
        tmp.append("for item, value in tmp.items():")
        tmp.append("    if isinstance(value, np.ndarray):")
        tmp.append("        tmp[item] = value.tolist()")
        tmp.append("")
        tmp.append("# Add the energy and gradients")
        tmp.append("tmp['energy'] = wfn.energy()")
        tmp.append("tmp['gradient'] = wfn.gradient().to_array().tolist()")
        tmp.append("")
        tmp.append("with open('final_structure.json', 'w') as fd:")
        tmp.append("    json.dump(tmp, fd, sort_keys=True, indent=3)")

        epilog = "\n".join(tmp)

        return prolog, epilog

    def run(self, node0, memory, n_threads):
        """Run the accelerated optimization.

        Parameters
        ----------
        node0 : seamm.Node
            The first node in the Psi4 subflowchart

        memory : str
            Amount of memory to allocate Psi4

        n_threads : int
            Number of threads to use in Psi4
        """
        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description = []
        self.description.append(__(self.description_text(PP), **PP, indent=self.indent))

        # Handle the system
        _, configuration = self.get_system_configuration(P)

        # Get the input other than the structure
        self.n_threads = n_threads
        self._node0 = node0
        self._memory = memory
        if self._restart_wfn is not None:
            self.prolog, self.epilog = self.input_text(
                node0, memory, restart="old_wfn.npy"
            )
        else:
            self.prolog, self.epilog = self.input_text(node0, memory)

        max_iterations = P["max geometry steps"]
        if max_iterations == "default":
            max_iterations = 2 * 3 * configuration.n_atoms

        self._counters["surrogate"] = 0
        self._counters["hessian"] = 0
        self._counters["main"] = 0

        # self._test(directory)
        # return

        use_surrogate = False
        if use_surrogate:
            if True:
                optimizer = Optimizer(
                    energy_callback=self.runPsi4, surrogate_callback=self.runMOPAC
                )
                x0 = configuration.atoms.get_coordinates()
                optimizer.optimize(x0)
            else:
                for self.iteration in range(max_iterations + 1):
                    converged = self._surrogate_optimization(
                        directory / f"step_{self.iteration:03d}"
                    )
                    if converged:
                        printer.normal("Converged!!!!!!")
                        break
                else:
                    printer.normal("NOT Converged!!!!!!")
        else:
            # Iterate
            # molecule = configuration.to_qcschema_dict()
            optking_options = {
                "step_type": P["optimization method"],
                "g_convergence": P["geometry convergence"],
                "opt_coordinates": P["coordinates"],
                "hess_update": P["hessian update"],
            }
            if P["recalc hessian"] == "every step":
                optking_options["full_hess_every"] = 1
            elif P["recalc hessian"] == "at beginning":
                optking_options["full_hess_every"] = 0
            elif P["recalc hessian"] == "never":
                optking_options["full_hess_every"] = -1
            else:
                optking_options["full_hess_every"] = P["recalc hessian"]

            # "max_force_g_convergence": 4.5e-04,
            # "rms_force_g_convergence": 3.0e-04,
            # "intrafrag_hess": "SIMPLE",

            # optimizer = optking.CustomHelper(molecule, optking_options)
            optimizer = None

            # The initial coordinates as NUMPY array. Psi4 may reorient, so need
            # to run and cache the first results
            for self.iteration in range(max_iterations + 1):
                subdirectory = directory / f"step_{self.iteration:03d}"
                subdirectory.mkdir(parents=True, exist_ok=True)

                needed = optimizer.calculations_needed()
                factor = Q_(1.0, "a_0").m_as("Å")
                xyz = []
                for row in optimizer.geom:
                    xyz.append([val * factor for val in row])

                configuration.atoms.set_coordinates(xyz)

                if "hessian" in needed:
                    # Run MOPAC
                    self._triangle = self._run_MOPAC_Hessian(subdirectory / "MOPAC")
                    H = np.array(self._hessian(None))
                    optimizer.HX = H

                E, dE = self._run_psi4(subdirectory)
                optimizer.E = E
                optimizer.gX = dE

                optimizer.compute()
                optimizer.take_step()

                lines = optimizer.test_convergence(str_mode="table").splitlines()
                if self.iteration == 0:
                    printer.normal("\n".join(lines[0:-2]))
                else:
                    printer.normal(lines[-3])
                converged = optimizer.test_convergence()
                if converged is True:
                    printer.normal("Optimization SUCCESS:")
                    break
            else:
                printer.normal("Optimization FAILURE:\n")

            result = optimizer.close()
            if False:
                printer.normal(result)

        printer.normal(f"   Calls to Psi4: {self._counters['main']:4d}")
        printer.normal(f"      to Hessian: {self._counters['hessian']:4d}")
        printer.normal(f"    to surrogate: {self._counters['surrogate']:4d}")

    def _test(self, directory):
        """Calculate derivatives as a check"""

        # Handle the system
        _, configuration = self.get_system_configuration()
        directory.mkdir(parents=True, exist_ok=True)

        xyz0 = configuration.atoms.get_coordinates(as_array=True)

        # Run MOPAC and Psi4
        Emopac, gradients = self._run_MOPAC_SPE(directory / "MOPAC")
        E0, dE0 = self._run_psi4(directory / "Psi4")

        print(f"{E0=:9.6f}")

        # And the correction to MOPAC dE
        delta_E = E0 - Emopac
        delta_dE = dE0 - gradients

        angstrom2bohr = Q_(1.0, "Å").m_as("a_0")
        direction = np.array(dE0) / norm(dE0)
        direction.reshape(configuration.n_atoms, 3)
        for factor in range(-3, 4, 1):
            dxyz = direction * (factor / 10)
            xyz = xyz0 + dxyz.reshape(configuration.n_atoms, 3) / angstrom2bohr
            configuration.atoms.set_coordinates(xyz)
            E, dE = self._run_MOPAC_SPE(directory / f"step{factor}")
            Ep, dEp = self._run_psi4(directory / f"psi4_{factor}")

            delta_E1 = np.dot(dxyz, delta_dE)
            E = E + delta_E + delta_E1
            dE += delta_dE

            print(
                f"{factor:2d} {E:9.6f} {np.dot(dE, direction.flatten()):9.6f} "
                f"{Ep:9.6f} {np.dot(dEp, direction.flatten()):9.6f}"
            )

    def _surrogate_optimization(self, directory):
        """Optimize using MOPAC gradients shifted by real ones."""
        debug = False

        # Handle the system
        _, configuration = self.get_system_configuration()

        # molecule = configuration.to_qcschema_dict()
        # # "g_convergence": "gau",
        # optking_options = {
        #     "opt_coordinates": "cartesian",
        #     "g_convergence": "cfour",
        #     "max_force_g_convergence": 4.5e-04,
        #     "rms_force_g_convergence": 3.0e-04,
        #     "FULL_HESS_EVERY": 1,
        #     "intrafrag_hess": "SIMPLE",
        # }
        # "hess_update": "none",
        # optimizer = optking.CustomHelper(molecule, optking_options)
        optimizer = None

        # First calculate the derivatives with both Psi4 and MOPAC
        xyz0 = np.array(optimizer.geom).flatten()
        factor = Q_(1.0, "a_0").m_as("Å")

        xyz = xyz0 * factor
        configuration.atoms.set_coordinates(xyz.reshape(configuration.n_atoms, 3))

        directory.mkdir(parents=True, exist_ok=True)

        # Run MOPAC and Psi4
        Emopac, gradients = self._run_MOPAC_SPE(directory / "MOPAC")
        E, dE = self._run_psi4(directory / "Psi4")

        tmp = gradients.reshape(configuration.n_atoms, 3)
        tmp2 = np.sum(tmp, axis=0)
        tmp3 = " ".join([f"{val:12.9f}" for val in tmp2])
        print(f"sum dEm {tmp3}")

        tmp = dE.reshape(configuration.n_atoms, 3)
        tmp2 = np.sum(tmp, axis=0)
        tmp3 = " ".join([f"{val:12.9f}" for val in tmp2])
        print(f"sum dEp {tmp3}")

        tmp -= tmp2 / configuration.n_atoms
        dE = tmp.flatten()

        tmp = dE.reshape(configuration.n_atoms, 3)
        tmp2 = np.sum(tmp, axis=0)
        tmp3 = " ".join([f"{val:12.9f}" for val in tmp2])
        print(f"sum dEp {tmp3}")

        if debug:
            print("MOPAC gradients")
            print(gradients)
            print("Psi4 derivatives")
            print(dE)

        # And the correction to MOPAC dE
        delta_E = E - Emopac
        delta_dE = dE - gradients

        if debug:
            m0 = abs_max(delta_dE)
            m1 = abs_max(gradients)
            m2 = abs_max(dE)
            rms2 = rms(dE)
            print(f"maxes: {m0:9.6f} {m1:9.6f} {m2:9.6f} {rms2:9.6f} {E:12.7f}")

        if debug:
            print("Correction")
            print(delta_dE)

        P = self.parameters.values_to_dict()
        max_iterations = P["max geometry steps"]
        if max_iterations == "default":
            max_iterations = 2 * 3 * configuration.n_atoms

        max_iterations = 2
        converged = False
        for self.iteration in range(max_iterations + 1):
            subdirectory = directory / f"step_{self.iteration:03d}"
            subdirectory.mkdir(parents=True, exist_ok=True)

            needed = optimizer.calculations_needed()
            xyz1 = np.array(optimizer.geom).flatten()
            factor = Q_(1.0, "a_0").m_as("Å")
            xyz = xyz1 * factor
            configuration.atoms.set_coordinates(xyz.reshape(configuration.n_atoms, 3))

            if "hessian" in needed:
                # Run MOPAC
                self._triangle = self._run_MOPAC_Hessian(subdirectory / "MOPAC")
                H = np.array(self._hessian(None))
                Unit = np.identity(configuration.n_atoms * 3) * 0.0
                optimizer.HX = H + Unit

                eigvals = np.linalg.eigvalsh(H + Unit).tolist()
                txt = " ".join([f"{val:6.3f}" for val in eigvals])
                print(f"Eigenvalues {txt}")

            E, dE = self._run_MOPAC_SPE(subdirectory)

            Esv = E
            dEsv = np.array(dE)

            if debug:
                # And Psi4 for testing
                Ex, dEx = self._run_psi4(subdirectory / "Psi4")

            dxyz = xyz1 - xyz0
            delta_E1 = np.dot(dxyz, delta_dE)

            if debug and optimizer.HX is not None:
                tmp = np.matmul(optimizer.HX, dxyz)
                m1 = tmp.max()
                m0 = delta_dE.max()
                e1 = np.dot(dxyz, tmp)
                e0 = np.dot(dxyz, delta_dE)
                print(f"max = {m1:9.6f} {m0:9.6f} delta E {e1:9.6f} {e0:9.6f}")

            E += delta_E + delta_E1
            dE += delta_dE

            if debug:
                m0 = abs_max(delta_dE)
                m1 = abs_max(dEsv)
                m2 = abs_max(dE)
                rms2 = rms(dE)
                print(f"-----> {m0:9.6f} {m1:9.6f} {m2:9.6f} {rms2:9.6f} {E:12.7f}")

            if debug:
                print(f"{Esv=} {delta_E=} {delta_E1=} --> {E} vs {Ex}")
                tmp = [f"{val:10.6f}" for val in dE.tolist()]
                print(" ".join(tmp))
                tmp = [f"{val:10.6f}" for val in dEx.tolist()]
                print(" ".join(tmp))

            optimizer.E = E
            optimizer.gX = dE

            optimizer.compute()
            optimizer.take_step()

            lines = optimizer.test_convergence(str_mode="table").splitlines()
            if self.iteration == 0:
                print("\n".join(lines[0:-2]))
            else:
                print(lines[-3])
            if optimizer.test_convergence():
                converged = self.iteration == 0

                result = optimizer.close()
                factor = Q_(1.0, "a_0").m_as("Å")
                xyzf = np.array(result["final_molecule"]["geometry"]) * factor
                # configuration.atoms.set_coordinates(
                #     xyzf.reshape(configuration.n_atoms, 3)
                # )

                tmp = " ".join([f"{val:9.6f}" for val in xyz.tolist()])
                print(f" xyz: {tmp}")
                tmp = " ".join([f"{val:9.6f}" for val in xyzf.tolist()])
                print(f"xyzf: {tmp}")
                break

        return converged

    def run_old(self, node0, memory, n_threads):
        """Run the accelerated optimization.

        Parameters
        ----------
        node0 : seamm.Node
            The first node in the Psi4 subflowchart

        memory : str
            Amount of memory to allocate Psi4

        n_threads : int
            Number of threads to use in Psi4
        """
        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description = []
        self.description.append(__(self.description_text(PP), **PP, indent=self.indent))

        # Handle the system
        _, configuration = self.get_system_configuration(P)

        # Get the input other than the structure
        self.n_threads = n_threads
        if self._restart_wfn is not None:
            self.prolog, self.epilog = self.input_text(
                node0, memory, restart="old_wfn.npy"
            )
        else:
            self.prolog, self.epilog = self.input_text(node0, memory)

        # Iterate
        max_iterations = P["max geometry steps"]
        if max_iterations == "default":
            max_iterations = 2 * 3 * configuration.n_atoms

        # The initial coordinates as NUMPY array. Psi4 may reorient, so need
        # to run and cache the first results
        self.iteration = 0

        x0 = configuration.atoms.get_coordinates(as_array=True).flatten()
        self._E, self._derivatives = self._step(x0)

        # Set the iteration count back this initial time
        self.iteration = 0

        subdirectory = directory / f"step_{self.iteration:03d}"
        json_file = subdirectory / "final_structure.json"
        if json_file.exists():
            with json_file.open() as fd:
                data = json.load(fd)
        else:
            raise RuntimeError("Structure and gradients from Psi4 not available.")

        self.iteration = 0
        x0 = np.array(data["geom"])
        configuration.atoms.set_coordinates(x0.reshape(configuration.n_atoms, 3))

        # Run MOPAC
        self._triangle = self._run_MOPAC_Hessian(subdirectory / "MOPAC")

        H = self._hessian(None)
        print("MOPAC Hessian")
        for n, row in enumerate(H.tolist()):
            txt = [f"{val:6.3f}" for val in row[: n + 1]]
            print(" ".join(txt))

        print("")

        singular = np.linalg.svd(H, compute_uv=False, hermitian=True).tolist()
        txt = [f"{val:.6f}" for val in singular]
        print("Singular values")
        print(" ".join(txt))
        print(f"{singular[-7]} {singular[-6]}")
        self._rcond = (singular[-7] + singular[-6]) / 2
        print("")

        Hk = np.linalg.pinv(H, rcond=self._rcond, hermitian=True)

        print("Inverse")
        for n, row in enumerate(Hk.tolist()):
            txt = [f"{val:6.3f}" for val in row[: n + 1]]
            print(" ".join(txt))

        print("")

        eigvals = np.linalg.eigvalsh(H).tolist()
        txt = [f"{val:6.3f}" for val in eigvals]
        print("Eigenvalues")
        print(" ".join(txt))
        print("")

        # Set up to restart from the previous wavefunction
        self.prolog, self.epilog = self.input_text(node0, memory, restart="old_wfn.npy")

        result = minimize(self._step, x0, method=self._bfgs, jac=True)
        # result = minimize(self._step, x0, method="BFGS", jac=True)
        # result = minimize(
        #     self._step,
        #     x0,
        #     method="trust-exact",
        #     jac=True,
        #     hess=self._hessian,
        # )

        # for key, value in result.items():
        #     print(key)
        #     print(value)
        print("BFGS inverse Hessian")
        for n, row in enumerate(result["hess_inv"].tolist()):
            txt = [f"{val:6.3f}" for val in row[: n + 1]]
            print(" ".join(txt))

        # Output the summary of the minimization
        if result["success"]:
            text = (
                "\nThe optimization converged in {nit} steps to {fun:.7f} E_h. "
                "This required {nfev} evaluations of the energy and derivatives."
            )
            printer.normal(__(text, **result, indent=4 * " "))
        else:
            text = (
                "\nThe optimization failed: {message} after {nit} steps, which "
                "required {nfev} evaluations of the energy and derivatives. The "
                "final energy was {fun:.7f} E_h."
            )

    def _run_MOPAC_SPE(self, directory):
        """Run the MOPAC flowchart in the given directory."""
        # Get the MOPAC flowchart
        self._counters["surrogate"] += 1
        directory.mkdir(parents=True, exist_ok=True)
        path = Path(pkg_resources.resource_filename(__name__, "data/"))
        flowchart = seamm.Flowchart(
            name="MOPAC",
            parent=self,
            namespace="org.molssi.seamm",
            directory=directory,
            parser_name="MOPAC subflowchart",
        )
        flowchart.read(path / "MOPAC_PM7_SPE.flow")

        # Find the handler for job.out and set the level up
        job_handler = None
        out_handler = None
        for handler in job.handlers:
            if (
                isinstance(handler, logging.FileHandler)
                and "job.out" in handler.baseFilename
            ):
                job_handler = handler
                job_level = job_handler.level
                job_handler.setLevel(printing.JOB)
            elif isinstance(handler, logging.StreamHandler):
                out_handler = handler
                out_level = out_handler.level
                out_handler.setLevel(printing.JOB)

        # Redirect output
        _file_handler = logging.FileHandler(directory / "subflowchart.out")
        _file_handler.setLevel(printing.NORMAL)
        formatter = logging.Formatter(fmt="{message:s}", style="{")
        _file_handler.setFormatter(formatter)
        job.addHandler(_file_handler)

        # Set up the argument parser for this node.
        parser = flowchart.parser

        parser.epilog = textwrap.dedent(
            """
            The plug-ins in this flowchart are listed above.
            Options, if any, for plug-ins are placed after
            the name of the plug-in, e.g.:

               test.flow lammps-step --log-level DEBUG --np 4

            To get help for a plug-in, use --help or -h after the
            plug-in name. E.g.

               test.flow lammps-step --help
            """
        )
        parser.usage = "%(prog)s [options] plug-in [options] plug-in [options] ..."

        # Now traverse the flowchart, setting up the ids and parsers
        flowchart.set_ids()
        flowchart.create_parsers()

        # And handle the command-line arguments and ini file options.
        parser.parse_args([])
        logger.info("Parsed the command-line arguments")

        # Get the command line options
        options = parser.get_options()

        # Set the options in each step
        for node in flowchart:
            step_type = node.step_type
            logger.info(f"    setting options for {step_type}")
            if step_type in options:
                node._options = options[step_type]
            if "SEAMM" in options:
                node._global_options = options["SEAMM"]

        # Write out an initial summary of the flowchart before doing anything
        # Reset the visited flag for traversal
        flowchart.reset_visited()

        # Get the start node
        next_node = flowchart.get_node("1")

        # describe ourselves
        printer.normal("\nDescription of the flowchart\n----------------------------")

        while next_node:
            # and print the description
            try:
                next_node = next_node.describe()
            except Exception:
                message = "Error describing the flowchart\n\n" + traceback.format_exc()
                print(message)
                logger.critical(message)
                raise
            except:  # noqa: E722
                message = (
                    "Unexpected error describing the flowchart\n\n"
                    + traceback.format_exc()
                )
                print(message)
                logger.critical(message)
                raise

        printer.normal("")

        # And actually run it!
        printer.normal(("Running the flowchart\n" "---------------------"))

        try:
            next_node = flowchart.get_node("1")
            while next_node is not None:
                try:
                    next_node = next_node.run()
                except DeprecationWarning as e:
                    print("\nDeprecation warning: " + str(e))
                    traceback.print_exc(file=sys.stderr)
                    traceback.print_exc(file=sys.stdout)
        finally:
            pass

        # Get the derivatives and return them.
        E = self.get_variable("_mopac_energy")
        gradients = self.get_variable("_mopac_gradients")
        dE = np.array(gradients).flatten()

        # Set output back
        _file_handler.close()
        job.removeHandler(_file_handler)
        if job_handler is not None:
            job_handler.setLevel(job_level)
        if out_handler is not None:
            out_handler.setLevel(out_level)

        return E, dE

    def _run_MOPAC_Hessian(self, directory):
        """Run the MOPAC flowchart in the given directory."""
        self._counters["hessian"] += 1
        # Get the MOPAC flowchart
        directory.mkdir(parents=True, exist_ok=True)
        path = Path(pkg_resources.resource_filename(__name__, "data/"))
        flowchart = seamm.Flowchart(
            name="MOPAC",
            parent=self,
            namespace="org.molssi.seamm",
            directory=directory,
            parser_name="MOPAC subflowchart",
        )
        flowchart.read(path / "MOPAC_PM7_Hessian.flow")

        # Find the handler for job.out and set the level up
        job_handler = None
        out_handler = None
        for handler in job.handlers:
            if (
                isinstance(handler, logging.FileHandler)
                and "job.out" in handler.baseFilename
            ):
                job_handler = handler
                job_level = job_handler.level
                job_handler.setLevel(printing.JOB)
            elif isinstance(handler, logging.StreamHandler):
                out_handler = handler
                out_level = out_handler.level
                out_handler.setLevel(printing.JOB)

        # Redirect output
        _file_handler = logging.FileHandler(directory / "subflowchart.out")
        _file_handler.setLevel(printing.NORMAL)
        formatter = logging.Formatter(fmt="{message:s}", style="{")
        _file_handler.setFormatter(formatter)
        job.addHandler(_file_handler)

        # Set up the argument parser for this node.
        parser = flowchart.parser

        parser.epilog = textwrap.dedent(
            """
            The plug-ins in this flowchart are listed above.
            Options, if any, for plug-ins are placed after
            the name of the plug-in, e.g.:

               test.flow lammps-step --log-level DEBUG --np 4

            To get help for a plug-in, use --help or -h after the
            plug-in name. E.g.

               test.flow lammps-step --help
            """
        )
        parser.usage = "%(prog)s [options] plug-in [options] plug-in [options] ..."

        # Now traverse the flowchart, setting up the ids and parsers
        flowchart.set_ids()
        flowchart.create_parsers()

        # And handle the command-line arguments and ini file options.
        parser.parse_args([])
        logger.info("Parsed the command-line arguments")

        # Get the command line options
        options = parser.get_options()

        # Set the options in each step
        for node in flowchart:
            step_type = node.step_type
            logger.info(f"    setting options for {step_type}")
            if step_type in options:
                node._options = options[step_type]
            if "SEAMM" in options:
                node._global_options = options["SEAMM"]

        # Write out an initial summary of the flowchart before doing anything
        # Reset the visited flag for traversal
        flowchart.reset_visited()

        # Get the start node
        next_node = flowchart.get_node("1")

        # describe ourselves
        printer.normal("\nDescription of the flowchart\n----------------------------")

        while next_node:
            # and print the description
            try:
                next_node = next_node.describe()
            except Exception:
                message = "Error describing the flowchart\n\n" + traceback.format_exc()
                print(message)
                logger.critical(message)
                raise
            except:  # noqa: E722
                message = (
                    "Unexpected error describing the flowchart\n\n"
                    + traceback.format_exc()
                )
                print(message)
                logger.critical(message)
                raise

        printer.normal("")

        # And actually run it!
        printer.normal(("Running the flowchart\n" "---------------------"))

        try:
            next_node = flowchart.get_node("1")
            while next_node is not None:
                try:
                    next_node = next_node.run()
                except DeprecationWarning as e:
                    print("\nDeprecation warning: " + str(e))
                    traceback.print_exc(file=sys.stderr)
                    traceback.print_exc(file=sys.stdout)
        finally:
            pass

        # Get the Hessian matrix and return it.
        triangle = []
        with open(directory / "1" / "hessian.dat") as fd:
            lines = fd.read().splitlines()

        if lines[0] != "!molssi hessian 1.0":
            raise IOError("Not a hessian file!")

        for line in lines[1:]:
            if line[0] == "@":
                pass
            else:
                triangle.append([float(val) for val in line.split()])

        # Set output back
        _file_handler.close()
        job.removeHandler(_file_handler)
        if job_handler is not None:
            job_handler.setLevel(job_level)
        if out_handler is not None:
            out_handler.setLevel(out_level)

        return triangle

    def _hessian(self, x, *args):
        """Return the Hessian matrix to SciPY."""
        n = len(self._triangle)
        matrix = np.zeros((n, n))
        i = 0
        for i, row in enumerate(self._triangle):
            for j, val in enumerate(row):
                matrix[i, j] = val
                matrix[j, i] = val

        return matrix

    def _run_psi4(self, directory):
        self._counters["main"] += 1
        directory.mkdir(parents=True, exist_ok=True)
        _, configuration = self.get_system_configuration()

        # Add the structure
        structure = self.parent._convert_structure(name="initial")

        # And run the calculation
        files = {}
        if self._restart_wfn is not None:
            files["old_wfn.npy"] = self._restart_wfn
            self.prolog, self.epilog = self.input_text(
                self._node0, self._memory, restart="old_wfn.npy"
            )
        else:
            self.prolog, self.epilog = self.input_text(self._node0, self._memory)
        files["input.dat"] = self.prolog + structure + self.epilog

        return_files = ["output.dat", "wfn.npy", "final_structure.json"]

        exe_path = Path(self.parent.options["psi4_path"])
        env = {
            "PSIPATH": str(exe_path),
            "PATH": str(exe_path),
        }

        local = seamm.ExecLocal()
        exe = exe_path / "psi4"
        result = local.run(
            cmd=[str(exe), f"-n {self.n_threads}"],
            files=files,
            return_files=return_files,
            env=env,
            in_situ=True,
            directory=directory,
        )

        if result is None:
            self.logger.error("There was an error running Psi4")
            return None

        # Capture the wavefunction for restarts
        self._restart_wfn = result["wfn.npy"]["data"]

        json_file = directory / "final_structure.json"
        if json_file.exists():
            with json_file.open() as fd:
                data = json.load(fd)
        else:
            raise RuntimeError("Structure and gradients from Psi4 not available.")

        self._E = data["energy"]
        self._derivatives = np.array(data["gradient"]).flatten()

        return self._E, self._derivatives

    def _step(self, x):
        _, configuration = self.get_system_configuration()

        x0 = configuration.atoms.get_coordinates(as_array=True).flatten()

        if self.iteration != 0 or self._E is None:
            # Update the coordinates in the configuration
            xyz = x.reshape(configuration.n_atoms, 3)
            configuration.atoms.set_coordinates(xyz)

            # Add the structure
            structure = self.parent._convert_structure(name="initial")

            # Make directory to work in
            directory = Path(self.directory)
            subdirectory = directory / f"step_{self.iteration:03d}"
            subdirectory.mkdir(parents=True, exist_ok=True)

            # And run the calculation
            files = {"input.dat": self.prolog + structure + self.epilog}

            if self._restart_wfn is not None:
                files["old_wfn.npy"] = self._restart_wfn

            return_files = ["output.dat", "wfn.npy", "final_structure.json"]

            exe_path = Path(self.parent.options["psi4_path"])
            env = {
                "PSIPATH": str(exe_path),
                "PATH": str(exe_path),
            }

            local = seamm.ExecLocal()
            exe = exe_path / "psi4"
            result = local.run(
                cmd=[str(exe), f"-n {self.n_threads}"],
                files=files,
                return_files=return_files,
                env=env,
                in_situ=True,
                directory=subdirectory,
            )

            if result is None:
                self.logger.error("There was an error running Psi4")
                return None

            # Capture the wavefunction for restarts
            self._restart_wfn = result["wfn.npy"]["data"]

            json_file = subdirectory / "final_structure.json"
            if json_file.exists():
                with json_file.open() as fd:
                    data = json.load(fd)
            else:
                raise RuntimeError("Structure and gradients from Psi4 not available.")

            self._E = data["energy"]
            self._derivatives = np.array(data["gradient"]).flatten()

            # Run MOPAC
            # self._triangle = self._run_MOPAC_Hessian(subdirectory / "MOPAC")

            if self.iteration == 0:
                text = " It    Energy (Ha)   RMS deriv  Max deriv  RMS xyz    Max xyz"
                printer.normal(__(text, indent=self.indent + 4 * " "))
                text = "--- --------------- ---------- ---------- ---------- ----------"
                printer.normal(__(text, indent=self.indent + 4 * " "))

            rmsd = np.sqrt(np.mean(np.square(self._derivatives)))
            maxd = np.amax(np.abs(self._derivatives))
            if self.iteration > 0:
                rms_step = np.sqrt(np.mean(np.square(x - x0)))
                max_step = np.amax(np.abs(x - x0))
                text = (
                    f"{self.iteration:3d} {self._E:15.7f} {rmsd:10.6f} {maxd:10.6f} "
                    f"{rms_step:10.6f} {max_step:10.6f}"
                )
            else:
                text = f"{self.iteration:3d} {self._E:15.7f} {rmsd:10.6f} {maxd:10.6f}"
            printer.normal(self.indent + 4 * " " + text)

        self.iteration += 1

        return self._E, self._derivatives

    def _bfgs(
        self,
        fun,
        x0,
        args=(),
        jac=None,
        callback=None,
        gtol=1e-5,
        norm=np.Inf,
        eps=_epsilon,
        maxiter=None,
        disp=False,
        return_all=False,
        finite_diff_rel_step=None,
        xrtol=0,
        **unknown_options,
    ):
        """
        Minimization of scalar function of one or more variables using the
        BFGS algorithm.

        Options
        -------
        disp : bool
            Set to True to print convergence messages.
        maxiter : int
            Maximum number of iterations to perform.
        gtol : float
            Terminate successfully if gradient norm is less than `gtol`.
        norm : float
            Order of norm (Inf is max, -Inf is min).
        eps : float or ndarray
            If `jac is None` the absolute step size used for numerical
            approximation of the jacobian via forward differences.
        return_all : bool, optional
            Set to True to return a list of the best solution at each of the
            iterations.
        finite_diff_rel_step : None or array_like, optional
            If `jac in ['2-point', '3-point', 'cs']` the relative step size to
            use for numerical approximation of the jacobian. The absolute step
            size is computed as ``h = rel_step * sign(x) * max(1, abs(x))``,
            possibly adjusted to fit into the bounds. For ``method='3-point'``
            the sign of `h` is ignored. If None (default) then step is selected
            automatically.
        xrtol : float, default: 0
            Relative tolerance for `x`. Terminate successfully if step size is
            less than ``xk * xrtol`` where ``xk`` is the current parameter vector.
        """
        # _check_unknown_options(unknown_options)
        retall = return_all

        x0 = np.asarray(x0).flatten()
        if x0.ndim == 0:
            x0.shape = (1,)
        if maxiter is None:
            maxiter = len(x0) * 200

        sf = _prepare_scalar_function(
            fun,
            x0,
            jac,
            args=args,
            epsilon=eps,
            finite_diff_rel_step=finite_diff_rel_step,
        )

        f = sf.fun
        myfprime = sf.grad

        old_fval = f(x0)
        gfk = myfprime(x0)

        k = 0
        N = len(x0)
        I = np.eye(N, dtype=int)  # noqa: E741

        H = self._hessian(None)
        Hk = np.linalg.pinv(H, rcond=self._rcond, hermitian=True)
        # Hk = I

        # Sets the initial step guess to dx ~ 1
        old_old_fval = old_fval + np.linalg.norm(gfk) / 2
        old_old_fval = None

        xk = x0
        if retall:
            allvecs = [x0]
        warnflag = 0
        gnorm = vecnorm(gfk, ord=norm)
        while (gnorm > gtol) and (k < maxiter):
            pk = -np.dot(Hk, gfk)
            print(f"{old_fval=} {old_old_fval=}")
            try:
                alpha_k, fc, gc, old_fval, old_old_fval, gfkp1 = _line_search_wolfe12(
                    f,
                    myfprime,
                    xk,
                    pk,
                    gfk,
                    old_fval,
                    old_old_fval,
                    amin=1e-100,
                    amax=1e100,
                )
            except _LineSearchError:
                # Line search failed to find a better solution.
                warnflag = 2
                break

            print(f"{old_fval=} {old_old_fval=} {alpha_k=}")
            sk = alpha_k * pk
            xkp1 = xk + sk

            if retall:
                allvecs.append(xkp1)
            xk = xkp1
            if gfkp1 is None:
                gfkp1 = myfprime(xkp1)

            yk = gfkp1 - gfk
            gfk = gfkp1
            if callback is not None:
                callback(xk)
            k += 1
            gnorm = vecnorm(gfk, ord=norm)
            if gnorm <= gtol:
                break

            #  See Chapter 5 in  P.E. Frandsen, K. Jonasson, H.B. Nielsen,
            #  O. Tingleff: "Unconstrained Optimization", IMM, DTU.  1999.
            #  These notes are available here:
            #  http://www2.imm.dtu.dk/documents/ftp/publlec.html
            if alpha_k * vecnorm(pk) <= xrtol * (xrtol + vecnorm(xk)):
                break

            if not np.isfinite(old_fval):
                # We correctly found +-Inf as optimal value, or something went
                # wrong.
                warnflag = 2
                break

            rhok_inv = np.dot(yk, sk)
            # this was handled in numeric, let it remaines for more safety
            # Cryptic comment above is preserved for posterity. Future reader:
            # consider change to condition below proposed in gh-1261/gh-17345.
            if rhok_inv == 0.0:
                rhok = 1000.0
                if disp:
                    print("Divide-by-zero encountered: rhok assumed large")
            else:
                rhok = 1.0 / rhok_inv

            A1 = I - sk[:, np.newaxis] * yk[np.newaxis, :] * rhok
            A2 = I - yk[:, np.newaxis] * sk[np.newaxis, :] * rhok
            Hk = np.dot(A1, np.dot(Hk, A2)) + (
                rhok * sk[:, np.newaxis] * sk[np.newaxis, :]
            )

        fval = old_fval

        if warnflag == 2:
            msg = _status_message["pr_loss"]
        elif k >= maxiter:
            warnflag = 1
            msg = _status_message["maxiter"]
        elif np.isnan(gnorm) or np.isnan(fval) or np.isnan(xk).any():
            warnflag = 3
            msg = _status_message["nan"]
        else:
            msg = _status_message["success"]

        if disp:
            print("%s%s" % ("Warning: " if warnflag != 0 else "", msg))
            print("         Current function value: %f" % fval)
            print("         Iterations: %d" % k)
            print("         Function evaluations: %d" % sf.nfev)
            print("         Gradient evaluations: %d" % sf.ngev)

        result = OptimizeResult(
            fun=fval,
            jac=gfk,
            hess_inv=Hk,
            nfev=sf.nfev,
            njev=sf.ngev,
            status=warnflag,
            success=(warnflag == 0),
            message=msg,
            x=xk,
            nit=k,
        )
        if retall:
            result["allvecs"] = allvecs
        return result

    def runPsi4(self, x, needed=("E", "dE")):
        self._counters["main"] += 1

        directory = Path(self.directory) / f"Psi4_{self._counters['main']}"
        directory.mkdir(parents=True, exist_ok=True)

        _, configuration = self.get_system_configuration()

        x_orig = configuration.atoms.get_coordinates()
        configuration.atoms.set_coordinates(x.reshape(configuration.n_atoms, 3))

        # Add the structure
        structure = self.parent._convert_structure(name="initial")

        configuration.atoms.set_coordinates(x_orig)

        # And run the calculation
        files = {}
        if self._restart_wfn is not None:
            files["old_wfn.npy"] = self._restart_wfn
            self.prolog, self.epilog = self.input_text(
                self._node0, self._memory, restart="old_wfn.npy"
            )
        else:
            self.prolog, self.epilog = self.input_text(self._node0, self._memory)
        files["input.dat"] = self.prolog + structure + self.epilog

        return_files = ["output.dat", "wfn.npy", "final_structure.json"]

        exe_path = Path(self.parent.options["psi4_path"])
        env = {
            "PSIPATH": str(exe_path),
            "PATH": str(exe_path),
        }

        local = seamm.ExecLocal()
        exe = exe_path / "psi4"
        result = local.run(
            cmd=[str(exe), f"-n {self.n_threads}"],
            files=files,
            return_files=return_files,
            env=env,
            in_situ=True,
            directory=directory,
        )

        if result is None:
            self.logger.error("There was an error running Psi4")
            return None

        # Capture the wavefunction for restarts
        self._restart_wfn = result["wfn.npy"]["data"]

        json_file = directory / "final_structure.json"
        if json_file.exists():
            with json_file.open() as fd:
                data = json.load(fd)
        else:
            raise RuntimeError("Structure and gradients from Psi4 not available.")

        E = data["energy"]
        dE = np.array(data["gradient"]).flatten() * Q_("E_h/a_0").m_as("E_h/Å")

        return E, dE

    def runMOPAC(self, x, needed=("E", "dE")):
        """Run the MOPAC flowchart in the given directory."""
        self._counters["surrogate"] += 1

        directory = Path(self.directory) / f"MOPAC_{self._counters['surrogate']}"
        directory.mkdir(parents=True, exist_ok=True)

        # Set the coordinates
        _, configuration = self.get_system_configuration()
        x_orig = configuration.atoms.get_coordinates()
        configuration.atoms.set_coordinates(x.reshape(configuration.n_atoms, 3))

        # Get the MOPAC flowchart
        path = Path(pkg_resources.resource_filename(__name__, "data/"))
        flowchart = seamm.Flowchart(
            name="MOPAC",
            parent=self,
            namespace="org.molssi.seamm",
            directory=directory,
            parser_name="MOPAC subflowchart",
        )
        flowchart.read(path / "MOPAC_PM7_surrogate.flow")

        # Find the handler for job.out and set the level up
        job_handler = None
        out_handler = None
        for handler in job.handlers:
            if (
                isinstance(handler, logging.FileHandler)
                and "job.out" in handler.baseFilename
            ):
                job_handler = handler
                job_level = job_handler.level
                job_handler.setLevel(printing.JOB)
            elif isinstance(handler, logging.StreamHandler):
                out_handler = handler
                out_level = out_handler.level
                out_handler.setLevel(printing.JOB)

        # Redirect output
        _file_handler = logging.FileHandler(directory / "subflowchart.out")
        _file_handler.setLevel(printing.NORMAL)
        formatter = logging.Formatter(fmt="{message:s}", style="{")
        _file_handler.setFormatter(formatter)
        job.addHandler(_file_handler)

        # Set up the argument parser for this node.
        parser = flowchart.parser

        parser.epilog = textwrap.dedent(
            """
            The plug-ins in this flowchart are listed above.
            Options, if any, for plug-ins are placed after
            the name of the plug-in, e.g.:

               test.flow lammps-step --log-level DEBUG --np 4

            To get help for a plug-in, use --help or -h after the
            plug-in name. E.g.

               test.flow lammps-step --help
            """
        )
        parser.usage = "%(prog)s [options] plug-in [options] plug-in [options] ..."

        # Now traverse the flowchart, setting up the ids and parsers
        flowchart.set_ids()
        flowchart.create_parsers()

        # And handle the command-line arguments and ini file options.
        parser.parse_args([])
        logger.info("Parsed the command-line arguments")

        # Get the command line options
        options = parser.get_options()

        # Set the options in each step
        for node in flowchart:
            step_type = node.step_type
            logger.info(f"    setting options for {step_type}")
            if step_type in options:
                node._options = options[step_type]
            if "SEAMM" in options:
                node._global_options = options["SEAMM"]

        # Write out an initial summary of the flowchart before doing anything
        # Reset the visited flag for traversal
        flowchart.reset_visited()

        # Get the start node
        next_node = flowchart.get_node("1")

        # describe ourselves
        printer.normal("\nDescription of the flowchart\n----------------------------")

        while next_node:
            # and print the description
            try:
                next_node = next_node.describe()
            except Exception:
                message = "Error describing the flowchart\n\n" + traceback.format_exc()
                print(message)
                logger.critical(message)
                raise
            except:  # noqa: E722
                message = (
                    "Unexpected error describing the flowchart\n\n"
                    + traceback.format_exc()
                )
                print(message)
                logger.critical(message)
                raise

        printer.normal("")

        # And actually run it!
        printer.normal(("Running the flowchart\n" "---------------------"))

        try:
            next_node = flowchart.get_node("1")
            while next_node is not None:
                try:
                    next_node = next_node.run()
                except DeprecationWarning as e:
                    print("\nDeprecation warning: " + str(e))
                    traceback.print_exc(file=sys.stderr)
                    traceback.print_exc(file=sys.stdout)
        finally:
            pass

        # Set the coordinates back
        configuration.atoms.set_coordinates(x_orig)

        # Get the derivatives and return them.
        E = self.get_variable("_mopac_energy")
        gradients = self.get_variable("_mopac_gradients")
        dE = np.array(gradients).flatten()

        # Get the Hessian matrix and return it.
        triangle = []
        with open(directory / "1" / "hessian.dat") as fd:
            lines = fd.read().splitlines()

        if lines[0] != "!molssi hessian 1.0":
            raise IOError("Not a hessian file!")

        for line in lines[1:]:
            if line[0] == "@":
                pass
            else:
                triangle.append([float(val) for val in line.split()])

        n = len(triangle)
        d2E = np.zeros((n, n))
        i = 0
        for i, row in enumerate(triangle):
            for j, val in enumerate(row):
                d2E[i, j] = val
                d2E[j, i] = val

        # Set output back
        _file_handler.close()
        job.removeHandler(_file_handler)
        if job_handler is not None:
            job_handler.setLevel(job_level)
        if out_handler is not None:
            out_handler.setLevel(out_level)

        if "d2E" in needed:
            return E, dE, d2E
        else:
            return E, dE



# -*- coding: utf-8 -*-

"""Acclerated optimization using MOPAC for Hessians."""

import logging

import numpy as np


def norm(V):
    return np.linalg.norm(V)


def abs_max(V):
    return max(abs(elem) for elem in V)


def abs_min(V):
    return min(abs(elem) for elem in V)


def rms(V):
    return np.sqrt(np.mean(V**2))


logger = logging.getLogger(__name__)


class Optimizer:
    def __init__(self, is_linear=False, energy_callback=None, surrogate_callback=None):
        self._E = None
        self._dE = None
        self._d2E = None
        self._iteration = 0
        self.is_linear = is_linear
        self._energy_callback = energy_callback
        self._surrogate_callback = surrogate_callback

    @property
    def E(self):
        """The energy."""
        return self._E

    @property
    def dE(self):
        """The derivatives of the energy."""
        return self._dE

    @property
    def d2E(self):
        """The second derivatives of the energy."""
        return self._d2E

    def optimize(self, _x0, n_steps=None):
        """Do the optimization!"""
        self.x0 = np.array(_x0).flatten()
        n_dof = len(self.x0)
        if self.is_linear:
            n_dof -= 5
        else:
            n_dof -= 6

        if n_steps is None:
            n_steps = 3 * n_dof

        x = np.array(self.x0)
        xlast = None
        print(
            f"{'Step':4} {'Energy':12} {'Max dE':9} {'rms dE':9} "
            f"{'Max step':9} {'rms step':9}"
        )
        print("---- ------------ --------- --------- --------- ---------")
        for step in range(0, n_steps + 1):
            if self._surrogate_callback is None:
                E, dE, d2E = self._energy_callback(x, ("E", "dE", "d2E"))
            else:
                Es, dEs, d2Es = self._surrogate_callback(x, ("E", "dE", "d2E"))
                E0, dE0 = self._energy_callback(x, ("E", "dE"))

                # Corrections to the surrogate energy and derivatives
                self.Ecorr = E0 - Es
                self.dEcorr = dE0 - dEs
                self.x0 = np.array(x)

                # Invert the Hessian carefully
                s = np.linalg.svd(d2Es, compute_uv=False, hermitian=True)
                if self.is_linear:
                    rcond = ((s[-6] + s[-5]) / 2) / s[0]
                else:
                    rcond = ((s[-7] + s[-6]) / 2) / s[0]

                A = np.linalg.pinv(d2Es, rcond=rcond)

                E = E0
                dE = np.array(dE0)

                Etmp = Es + self.Ecorr + np.dot(x - self.x0, self.dEcorr)
                dEtmp = dEs + self.dEcorr

                tmp = " ".join([f"{v:9.6f}" for v in dEtmp.tolist()])
                print("")
                print(f"MOPAC {Etmp:12.7f}  x: {tmp}")
                tmp = " ".join([f"{v:9.6f}" for v in dE.tolist()])
                print(f" Psi4 {E:12.7f}  x: {tmp}")
                print("")

            max_dE = abs_max(dE)
            rms_dE = rms(dE)
            if step == 0:
                print(f"{step:4d} {E:12.7f} {max_dE:9.6f} {rms_dE:9.6f}")
            else:
                delta = xlast - x
                max_dx = abs_max(delta)
                rms_dx = rms(delta)
                print(
                    f"{step:4d} {E:12.7f} {max_dE:9.6f} {rms_dE:9.6f} "
                    f"{max_dx:9.6f} {rms_dx:9.6f}"
                )

            xlast = np.array(x)

            if True:
                x = self._optimize(x, E, dE, A)
            else:
                direction = np.matmul(A, -dE)
                stepsize, Enew, x = self._linesearch(
                    x, E, dE, direction, self._surrogate
                )

    def _optimize(self, x, E, dE, A):
        n_dof = len(self.x0)
        if self.is_linear:
            n_dof -= 5
        else:
            n_dof -= 6
        n_steps = 3 * n_dof
        xlast = None
        for step in range(0, n_steps + 1):
            tmp = " ".join([f"{v:9.6f}" for v in x.tolist()])
            logger.debug(f"coordinates: {tmp}")

            tmp = " ".join([f"{v:9.6f}" for v in x.tolist()])
            print(f"\t\t\t\t\t         x: {tmp}")

            if step >= 0:
                E, dE, d2E = self._surrogate(x, ("E", "dE", "d2E"))

                # Invert the Hessian carefully
                s = np.linalg.svd(d2E, compute_uv=False, hermitian=True)
                if self.is_linear:
                    rcond = ((s[-6] + s[-5]) / 2) / s[0]
                else:
                    rcond = ((s[-7] + s[-6]) / 2) / s[0]

                A = np.linalg.pinv(d2E, rcond=rcond, hermitian=True)

            if False and step > 0:
                E0, dE0 = self._energy_callback(x, ("E", "dE"))
                E = E0
                dE = dE0

                tmp = " ".join([f"{v:9.6f}" for v in dE.tolist()])
                print(f"\t\t\t\t\tMOPAC {E:12.7f}  x: {tmp}")
                tmp = " ".join([f"{v:9.6f}" for v in dE0.tolist()])
                print(f"\t\t\t\t\t Psi4 {E0:12.7f}  x: {tmp}")

            tmp = " ".join([f"{v:9.6f}" for v in dE.tolist()])
            logger.debug(f"         dE: {tmp}")

            max_dE = abs_max(dE)
            rms_dE = rms(dE)
            if step == 0:
                print(f"\t{step:4d} {E:12.7f} {max_dE:9.6f} {rms_dE:9.6f}")
            else:
                delta = xlast - x
                max_dx = abs_max(delta)
                rms_dx = rms(delta)
                print(
                    f"\t{step:4d} {E:12.7f} {max_dE:9.6f} {rms_dE:9.6f} "
                    f"{max_dx:9.6f} {rms_dx:9.6f}"
                )

            if max_dE < 1.0e-04 and rms_dE < 1.0e-04:
                break

            xlast = np.array(x)

            direction = np.matmul(A, -dE)
            # direction = -dE

            tmp = " ".join([f"{v:9.6f}" for v in dE.tolist()])
            print(f"\t\t\t\t\t          dE: {tmp}")
            tmp = " ".join([f"{v:9.6f}" for v in direction.tolist()])
            print(f"\t\t\t\t\t   direction: {tmp}")

            direction, residuals, rank, s = np.linalg.lstsq(d2E, -dE, rcond=rcond)

            print(f"{rank=}")
            tmp = " ".join([f"{v:9.6f}" for v in direction.tolist()])
            print(f"\t\t\t\t\t   direction: {tmp}")

            stepsize, Enew, x = self._linesearch(x, E, dE, direction, self._surrogate)
            logger.debug(f"stepsize = {stepsize:.4f}")
        logger.debug("returning x")
        logger.debug(x)
        tmp = " ".join([f"{v:9.6f}" for v in x.tolist()])
        print(f"\t\t\t\t\t         x: {tmp}")
        return x

    def _linesearch(self, x0, E0, dE0, direction, callback):
        stepsize = 1.0
        dE0line = np.dot(dE0, direction)
        while True:
            x = x0 + direction * stepsize
            E, dE = self._surrogate(x)
            dEline = np.dot(dE, direction)
            logger.debug(
                f"{stepsize=} {E0:9.6f} --> {E:9.6f}, {dE0line:9.6f} --> {dEline:9.6f}"
            )
            if abs(dEline) < 1.0e-8:
                break

            a = abs(dE0line)
            b = abs(dEline)
            if dE0line < 0:
                if dEline > 0:
                    if a > b:
                        delta = 1 - b / a
                    else:
                        delta = a / b
                else:
                    if b > a:
                        # Stepped over a maximum?
                        raise RuntimeError("Y > X stepped over a maximum?")
                    else:
                        delta = 1 + b / a
            else:
                raise RuntimeError("Line search is backwards?")
            logger.debug(f"{delta=:9.6f}")
            if abs(delta - 1) < 0.1:
                break
            stepsize *= delta
        return stepsize, E, x

    def _surrogate(self, x, needed=("E", "dE")):
        if "d2E" in needed:
            Es, dEs, d2E = self._surrogate_callback(x, needed)
        else:
            Es, dEs = self._surrogate_callback(x, needed)

        E = Es + self.Ecorr + np.dot(x - self.x0, self.dEcorr)
        dE = dEs + self.dEcorr

        if "d2E" in needed:
            return E, dE, d2E
        else:
            return E, dE



# -*- coding: utf-8 -*-

"""Non-graphical part of the Psi4 step in a SEAMM flowchart
"""

import configparser
import importlib
import json
import logging
from pathlib import Path
import pprint
import shutil

import psi4_step
import seamm
import seamm_exec
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: 'job' and 'printer'. 'job' send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# 'printer' sends output to the file 'step.out' in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("Psi4")

pre_code = """\
def fix_multipoles(data):
    result = {}
    esp = []
    it = iter(data.items())
    for key, value in it:
        if 'PROP ' == key[0:5]:
            result[key[6:]] = value[0]
        elif 'SCF DIPOLE' == key:
            result[key] = value[0]
        elif 'CURRENT DIPOLE' == key:
            result[key] = value[0]
        elif '32-POLE' in key:
            tmp = []
            while True:
                value = 0.0 if abs(value) < 1.0e-10 else value
                tmp.append(value)
                if 'ZZZZZ' in key:
                    break
                key, value = next(it)
            result['32-POLE'] = tmp
        elif 'HEXADECAPOLE' in key:
            tmp = []
            while True:
                value = 0.0 if abs(value) < 1.0e-10 else value
                tmp.append(value)
                if 'ZZZZ' in key:
                    break
                key, value = next(it)
            result['HEXADECAPOLE'] = tmp
        elif 'OCTUPOLE' in key:
            tmp = []
            while True:
                value = 0.0 if abs(value) < 1.0e-10 else value
                tmp.append(value)
                if 'ZZZ' in key:
                    break
                key, value = next(it)
            result['OCTUPOLE'] = tmp
        elif 'QUADRUPOLE' in key:
            tmp = []
            while True:
                value = 0.0 if abs(value) < 1.0e-10 else value
                tmp.append(value)
                if 'ZZ' in key:
                    break
                key, value = next(it)
            result['QUADRUPOLE'] = tmp
        elif 'DIPOLE' in key:
            tmp = []
            while True:
                value = 0.0 if abs(value) < 1.0e-10 else value
                tmp.append(value)
                result[key] = value
                if 'Z' in key:
                    break
                key, value = next(it)
            result[key[0:-2]] = tmp
        elif 'ESP AT CENTER' in key:
            esp.append(value)
        else:
            result[key] = value

    if len(esp) > 0:
        result['ELECTROSTATIC POTENTIAL'] = esp
    return result
"""


def humanize(memory, suffix="B", kilo=1024):
    """
    Scale memory to its proper format e.g:

        1253656 => '1.20 MiB'
        1253656678 => '1.17 GiB'
    """
    if kilo == 1000:
        units = ["", "k", "M", "G", "T", "P"]
    elif kilo == 1024:
        units = ["", "Ki", "Mi", "Gi", "Ti", "Pi"]
    else:
        raise ValueError("kilo must be 1000 or 1024!")

    for unit in units:
        if memory < kilo:
            return f"{memory:.2f} {unit}{suffix}"
        memory /= kilo


def dehumanize(memory, suffix="B"):
    """
    Unscale memory from its human readable form e.g:

        '1.20 MB' => 1200000
        '1.17 GB' => 1170000000
    """
    units = {
        "": 1,
        "k": 1000,
        "M": 1000**2,
        "G": 1000**3,
        "P": 1000**4,
        "Ki": 1024,
        "Mi": 1024**2,
        "Gi": 1024**3,
        "Pi": 1024**4,
    }

    tmp = memory.split()
    if len(tmp) == 1:
        return memory
    elif len(tmp) > 2:
        raise ValueError("Memory must be <number> <units>, e.g. 1.23 GB")

    amount, unit = tmp
    amount = float(amount)

    for prefix in units:
        if prefix + suffix == unit:
            return int(amount * units[prefix])

    raise ValueError(f"Don't recognize the units on '{memory}'")


class Psi4(seamm.Node):
    """
    The non-graphical part of a Psi4 step in a flowchart.

    Attributes
    ----------
    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : Psi4Parameters
        The control parameters for Psi4.

    See Also
    --------
    TkPsi4,
    Psi4, Psi4Parameters
    """

    def __init__(
        self,
        flowchart=None,
        title="Psi4",
        namespace="org.molssi.seamm.psi4",
        extension=None,
    ):
        """A step for Psi4 in a SEAMM flowchart.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
            flowchart: seamm.Flowchart
                The non-graphical flowchart that contains this step.

            title: str
                The name displayed in the flowchart.
            namespace: The namespace for the plugins of the subflowchart
            extension: None
                Not yet implemented
        """
        logger.debug("Creating Psi4 {}".format(self))

        self.subflowchart = seamm.Flowchart(
            parent=self, name="Psi4", namespace=namespace
        )

        super().__init__(
            flowchart=flowchart, title="Psi4", extension=extension, logger=logger
        )

        self.parameters = psi4_step.Psi4Parameters()

    @property
    def version(self):
        """The semantic version of this module."""
        return psi4_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return psi4_step.__git_revision__

    def create_parser(self):
        """Setup the command-line / config file parser"""
        # parser_name = 'psi4-step'
        parser_name = self.step_type
        parser = self.flowchart.parser

        # Remember if the parser exists ... this type of step may have been
        # found before
        parser_exists = parser.exists(parser_name)

        # Create the standard options, e.g. log-level
        result = super().create_parser(name=parser_name)

        if parser_exists:
            return result

        # Options for Psi4
        parser.add_argument(
            parser_name,
            "--ncores",
            default="available",
            help="How many threads to use in Psi4",
        )

        parser.add_argument(
            parser_name,
            "--memory",
            default="available",
            help=(
                "The maximum amount of memory to use for Psi4, which can be "
                "'all' or 'available', or a number, which may use k, Ki, "
                "M, Mi, etc. suffixes. Default: available."
            ),
        )

        parser.add_argument(
            parser_name,
            "--memory-factor",
            default="90%",
            help=(
                "The amount of possible memory to use, typically about 90%% to"
                " allow for Psi4 not keeping track of all the memory used."
            ),
        )

        return result

    def set_id(self, node_id):
        """Set the id for node to a given tuple"""
        self._id = node_id

        # and set our subnodes
        self.subflowchart.set_ids(self._id)

        return self.next()

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
            P: dict
                An optional dictionary of the current values of the control
                parameters.
        Returns
        -------
            description : str
                A description of the current step.
        """

        self.subflowchart.root_directory = self.flowchart.root_directory

        # Get the first real node
        node = self.subflowchart.get_node("1").next()

        text = self.header + "\n\n"
        while node is not None:
            text += __(node.description_text(), indent=4 * " ").__str__()
            text += "\n"
            node = node.next()

        return text

    def run(self):
        """Run a Psi4 step.

        Returns
        -------

        next_node : seamm.Node
            The next node object in the flowchart.

        """
        printer.important(self.header)
        printer.important("")

        # Add the main citation for Psi4
        self.references.cite(
            raw=self._bibliography["doi:10.1063/5.0006002"],
            alias="psi4",
            module="psi4 step",
            level=1,
            note="The principle Psi4 citation.",
        )

        system_db = self.get_variable("_system_db")
        configuration = system_db.system.configuration
        n_atoms = configuration.n_atoms
        if n_atoms == 0:
            self.logger.error("Psi4 run(): there is no structure!")
            raise RuntimeError("Psi4 run(): there is no structure!")

        # Access the options
        options = self.options
        seamm_options = self.global_options

        # Get the computational environment and set limits
        ce = seamm_exec.computational_environment()

        # Maximum number of threads
        n_threads = ce["NTASKS"]
        if options["ncores"] == "available":
            pass
        else:
            n_threads = int(options["ncores"])
            if n_threads < 1:
                n_threads = 1
        if seamm_options["ncores"] != "available":
            tmp = int(seamm_options["ncores"])
            if tmp < n_threads:
                n_threads = tmp
        ce["NTASKS"] = n_threads

        # And memory
        if seamm_options["memory"] == "all":
            mem_limit = ce["MEM_PER_NODE"]
        elif seamm_options["memory"] == "available":
            # For the default, 'available', use in proportion to number of
            # cores used
            mem_limit = n_threads * ce["MEM_PER_CPU"]
        else:
            mem_limit = dehumanize(seamm_options["memory"])

        if options["memory"] == "all":
            memory = ce["MEM_PER_NODE"]
        elif options["memory"] == "available":
            # For the default, 'available', use in proportion to number of
            # cores used
            memory = n_threads * ce["MEM_PER_CPU"]
        else:
            memory = dehumanize(options["memory"])

        memory = min(memory, mem_limit)

        if "%" in options["memory_factor"]:
            factor = float(options["memory_factor"].rstrip("%")) / 100.0
        else:
            factor = float(options["memory_factor"])

        memory *= factor

        # Psi applies a minimum of 250 MiB
        min_memory = dehumanize("250 MiB")
        if min_memory > memory:
            memory = min_memory
        ce["MEM_PER_NODE"] = memory
        memory = humanize(memory, kilo=1024)

        # Work through the subflowchart to find out what to do.
        self.subflowchart.root_directory = self.flowchart.root_directory

        next_node = super().run(printer)

        # Get the first real node
        node0 = self.subflowchart.get_node("1").next()

        # Start the input data
        input_data = []
        input_data.append("import json")
        input_data.append("import numpy as np")
        input_data.append("from pathlib import Path")
        input_data.append("import pprint")
        input_data.append("")
        input_data.append(f"memory {memory}")
        input_data.append("")
        input_data.append(pre_code)
        input_data.append("")

        # Put the structure into the input
        input_data.append(self._convert_structure(name="initial"))

        node = node0
        while node is not None:
            text = node.get_input()
            input_data.append(text)

            input_data.append("clean()")
            input_data.append("clean_variables()")
            # input_data.append('clean_options()')

            node = node.next()

        # Write out the final structure
        input_data.append("")
        input_data.append("# Write the final structure to disk")
        input_data.append("molecule = get_active_molecule()")
        input_data.append("tmp = molecule.to_dict()")
        input_data.append("for item, value in tmp.items():")
        input_data.append("    if isinstance(value, np.ndarray):")
        input_data.append("        tmp[item] = value.tolist()")
        input_data.append("")
        input_data.append("with open('final_structure.json', 'w') as fd:")
        input_data.append("    json.dump(tmp, fd, sort_keys=True, indent=3)")

        files = {"input.dat": "\n".join(input_data)}
        self.logger.info("input.dat:\n" + files["input.dat"])

        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        return_files = ["output.dat", "*.json", "*.cube"]

        # Check for already having run
        path = Path(self.directory) / "success.dat"
        if path.exists():
            result = {}
            path = Path(self.directory) / "stdout.txt"
            if path.exists():
                result["stdout"] = path.read_text()
            result["stderr"] = ""
            failed = False
        else:
            executor = self.flowchart.executor

            # Read configuration file for Psi4 if it exists
            executor_type = executor.name
            full_config = configparser.ConfigParser()
            ini_dir = Path(seamm_options["root"]).expanduser()
            path = ini_dir / "psi4.ini"

            if path.exists():
                full_config.read(ini_dir / "psi4.ini")

            # If the section we need doesn't exists, get the default
            if not path.exists() or executor_type not in full_config:
                resources = importlib.resources.files("psi4_step") / "data"
                ini_text = (resources / "psi4.ini").read_text()
                full_config.read_string(ini_text)

            # Getting desperate! Look for an executable in the path
            if executor_type not in full_config:
                path = shutil.which("psi4")
                if path is None:
                    raise RuntimeError(
                        f"No section for '{executor_type}' in Psi4 ini file "
                        f"({ini_dir / 'psi4.ini'}), nor in the defaults, nor "
                        "in the path!"
                    )
                else:
                    full_config[executor_type] = {
                        "installation": "local",
                        "code": f"{path} -n {{NTASKS}}",
                    }

            config = dict(full_config.items(executor_type))
            # Use the matching version of the seamm-psi4 image by default.
            config["version"] = self.version

            printer.important(
                self.indent
                + f"    Psi4 will use {ce['NTASKS']} threads and {memory} memory\n"
            )

            result = executor.run(
                cmd=["{code}"],
                config=config,
                directory=self.directory,
                files=files,
                return_files=return_files,
                in_situ=True,
                shell=True,
                ce=ce,
            )

            if not result:
                self.logger.error("There was an error running Psi4")
                raise RuntimeError("There was an error running Psi4")

            self.logger.debug("\n" + pprint.pformat(result))

            failed = False
            if "output.dat" in result["files"]:
                if result["output.dat"]["data"] is not None:
                    if (
                        "*** Psi4 exiting successfully."
                        not in result["output.dat"]["data"]
                    ):
                        self.logger.warning("Psi4 did not complete successfully.")
                        failed = True
            if not failed:
                # Write a small file to say that LAMMPS ran successfully, so cancel
                # skip if rerunning.
                path = Path(self.directory) / "success.dat"
                path.write_text("success")

        # Analyze the results
        self.analyze()

        if failed:
            raise RuntimeError("Psi4 did not complete successfully.")

        return next_node

    def analyze(self, indent="", **kwargs):
        """Do any analysis of the output from this step.

        Also print important results to the local step.out file using
        'printer'.

        Parameters
        ----------
            indent: str
                An extra indentation for the output
        """
        # Update the structure
        directory = Path(self.directory)
        structure_file = directory / "final_structure.json"
        if structure_file.exists():
            with structure_file.open(mode="r") as fd:
                structure = json.load(fd)
            if "geom" in structure:
                system_db = self.get_variable("_system_db")
                configuration = system_db.system.configuration
                xs = []
                ys = []
                zs = []
                it = iter(structure["geom"])
                for x in it:
                    xs.append(x)
                    ys.append(next(it))
                    zs.append(next(it))
                configuration.atoms["x"][0:] = xs
                configuration.atoms["y"][0:] = ys
                configuration.atoms["z"][0:] = zs
                printer.important(
                    self.indent + "    Updated the system with the structure from Psi4",
                )
                printer.important("")

        # Get the first real node
        node = self.subflowchart.get_node("1").next()

        # Loop over the subnodes, asking them to do their analysis
        while node is not None:
            for value in node.description:
                printer.important(value)

            node.analyze()

            printer.normal("")

            node = node.next()

    def _convert_structure(self, name=None, no_com=True, no_reorient=True):
        """Convert the structure to the input for Psi4."""

        system_db = self.get_variable("_system_db")
        configuration = system_db.system.configuration

        structure = []
        if name is None:
            structure.append("molecule {")
        else:
            structure.append("molecule " + name + " {")

        # Charge and multiplicity
        structure.append(f"  {configuration.charge} {configuration.spin_multiplicity}")

        elements = configuration.atoms.symbols
        coordinates = configuration.atoms.coordinates

        if "freeze" in configuration.atoms:
            freeze = configuration.atoms["freeze"]
        else:
            freeze = [""] * len(elements)

        for element, xyz, frz in zip(elements, coordinates, freeze):
            x, y, z = xyz
            structure.append(
                f"    {element:2s} {float(x): 12.8f} {float(y): 12.8f} "
                f"{float(z): 12.8f}"
            )
        structure.append("")
        if no_com:
            structure.append("no_com")
        if no_reorient:
            structure.append("no_reorient")
        structure.append("}")

        return "\n".join(structure) + "\n"



# -*- coding: utf-8 -*-

"""The graphical part of a Thermochemistry step"""

import logging
import tkinter as tk
import tkinter.ttk as ttk

import psi4_step
import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkThermochemistry(psi4_step.TkEnergy):
    """
    The graphical part of a Thermochemistry step in a flowchart.

    Attributes
    ----------
    tk_flowchart : TkFlowchart = None
        The flowchart that we belong to.
    node : Node = None
        The corresponding node of the non-graphical flowchart
    canvas: tkCanvas = None
        The Tk Canvas to draw on
    dialog : Dialog
        The Pmw dialog object
    x : int = None
        The x-coordinate of the center of the picture of the node
    y : int = None
        The y-coordinate of the center of the picture of the node
    w : int = 200
        The width in pixels of the picture of the node
    h : int = 50
        The height in pixels of the picture of the node
    self[widget] : dict
        A dictionary of tk widgets built using the information
        contained in Thermochemistry_parameters.py

    See Also
    --------
    Thermochemistry, TkThermochemistry,
    ThermochemistryParameters,
    """

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50,
        my_logger=logger,
    ):
        """
        Initialize a graphical node.

        Parameters
        ----------
        tk_flowchart: Tk_Flowchart
            The graphical flowchart that we are in.
        node: Node
            The non-graphical node for this step.
        namespace: str
            The stevedore namespace for finding sub-nodes.
        canvas: Canvas
           The Tk canvas to draw on.
        x: float
            The x position of the nodes center on the canvas.
        y: float
            The y position of the nodes cetner on the canvas.
        w: float
            The nodes graphical width, in pixels.
        h: float
            The nodes graphical height, in pixels.

        Returns
        -------
        None
        """
        self.dialog = None

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
        )

    def create_dialog(self):
        """
        Create the dialog. A set of widgets will be chosen by default
        based on what is specified in the Thermochemistry_parameters
        module.

        Parameters
        ----------
        None

        Returns
        -------
        None

        See Also
        --------
        TkThermochemistry.reset_dialog
        """

        frame = super().create_dialog(title="Thermochemistry")

        # Shortcut for parameters
        P = self.node.parameters

        # Frame to isolate widgets
        thermo_frame = self["thermochemistry"] = ttk.LabelFrame(
            frame,
            borderwidth=4,
            relief="sunken",
            text="Thermochemistry",
            labelanchor="n",
            padding=10,
        )
        # Then create the widgets
        for key in psi4_step.ThermochemistryParameters.parameters:
            self[key] = P[key].widget(thermo_frame)

        # and binding to change as needed
        for key in ("use existing parameters",):
            self[key].combobox.bind("<<ComboboxSelected>>", self.reset_thermochemistry)
            self[key].combobox.bind("<Return>", self.reset_thermochemistry)
            self[key].combobox.bind("<FocusOut>", self.reset_thermochemistry)

        # Top level needs to call reset_dialog
        if self.node.calculation == "thermochemistry":
            self.reset_dialog()

        return frame

    def reset_dialog(self, widget=None, row=0):
        """Layout the widgets in the dialog.

        The widgets are chosen by default from the information in
        Thermochemistry_parameter.

        This function simply lays them out row by row with
        aligned labels. You may wish a more complicated layout that
        is controlled by values of some of the control parameters.
        If so, edit or override this method

        Parameters
        ----------
        widget : Tk Widget = None

        Returns
        -------
        None

        See Also
        --------
        TkThermochemistry.create_dialog
        """
        # Remove any widgets previously packed
        frame = self["frame"]
        if row == 0:
            for slave in frame.grid_slaves():
                slave.grid_forget()

        # Shortcut for parameters
        P = self.node.parameters

        self["thermochemistry"].grid(row=row, column=0)
        row += 1

        self.reset_thermochemistry()

        if not P["use existing parameters"]:
            row = super().reset_dialog(row=row)
        else:
            self.reset_calculation()
        return row

    def reset_thermochemistry(self, widget=None):
        frame = self["thermochemistry"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        widgets = []
        row = 0

        for key in ("use existing parameters", "T", "P"):
            self[key].grid(row=row, column=0, sticky=tk.EW)
            widgets.append(self[key])
            row += 1

        sw.align_labels(widgets)



# -*- coding: utf-8 -*-

import psi4_step


class Psi4Step(object):
    """Helper class needed for the stevedore integration.

    This must provide a description() method that returns a dict containing a
    description of this node, and create_node() and create_tk_node() methods
    for creating the graphical and non-graphical nodes.

    Parameters
    ----------
    my_description : {description, group, name}
        A human-readable description of this step. It can be
        several lines long, and needs to be clear to non-expert users.
        It contains the following keys: description, group, name.
    my_description['description'] : tuple
        A description of the Psi4 step. It must be
        clear to non-experts.
    my_description['group'] : str
        Which group in the menus to put this step. If the group does
        not exist it will be created. Common groups are 'Building',
        'Calculations', 'Control' and 'Data'.
    my_description['name'] : str
        The name of this step, to be displayed in the menus.
    """

    my_description = {
        "description": ("An interface for Psi4"),
        "group": "Simulations",
        "name": "Psi4",
    }

    def __init__(self, flowchart=None, gui=None):
        """Initialize this helper class, which is used by
        the application via stevedore to get information about
        and create node objects for the flowchart
        """
        pass

    def description(self):
        """Return a description of what this extension does"""
        return Psi4Step.my_description

    def create_node(self, flowchart=None, **kwargs):
        """Create and return the new node object.

        Parameters
        ----------
        flowchart: seamm.Node
            A non-graphical SEAMM node

        **kwargs : keyworded arguments
            Various keyworded arguments such as title, namespace or
            extension representing the title displayed in the flowchart,
            the namespace for the plugins of a subflowchart and
            the extension, respectively.

        Returns
        -------
        Psi4

        See Also
        --------
        Psi4

        """

        return psi4_step.Psi4(flowchart=flowchart, **kwargs)

    def create_tk_node(self, canvas=None, **kwargs):
        """Create and return the graphical Tk node object.

        Parameters
        ----------
        canvas : tk.Canvas
            The Tk Canvas widget

        **kwargs : keyworded arguments
            Various keyworded arguments such as tk_flowchart, node, x, y, w, h
            representing a graphical flowchart object, a non-graphical node for
            a step, and dimensions of the graphical node.

        Returns
        -------
        TkPsi4

        See Also
        --------
        TkPsi4
        """

        return psi4_step.TkPsi4(canvas=canvas, **kwargs)



# -*- coding: utf-8 -*-

"""Setup and run Psi4"""

import json
import logging
from pathlib import Path

from openbabel import openbabel

import psi4_step
import seamm
import seamm.data
from seamm_util import units_class
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("psi4")


class Energy(seamm.Node):
    def __init__(self, flowchart=None, title="Energy", extension=None, logger=logger):
        """Initialize the node"""

        logger.debug("Creating Energy {}".format(self))

        super().__init__(
            flowchart=flowchart, title=title, extension=extension, logger=logger
        )

        self._calculation = "energy"
        self._model = None
        self._metadata = psi4_step.metadata
        self.parameters = psi4_step.EnergyParameters()

        self.description = "A single point energy calculation"

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return psi4_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return psi4_step.__git_revision__

    def description_text(
        self,
        P=None,
        calculation_type="Single-point energy",
        configuration=None,
    ):
        """Prepare information about what this node will do"""

        if P is None:
            P = self.parameters.values_to_dict()

        if P["level"] == "recommended":
            method = P["method"]
        else:
            method = P["advanced_method"]

        if self.is_expr(method):
            text = f"{calculation_type} using method given by {method}."
        elif (
            method in psi4_step.methods and psi4_step.methods[method]["method"] == "dft"
        ):
            if P["level"] == "recommended":
                functional = P["functional"]
            else:
                functional = P["advanced_functional"]
            text = f"{calculation_type} using {method} with an "
            text += f"exchange-correlation potential of {functional}"
            if (
                functional in psi4_step.dft_functionals
                and len(psi4_step.dft_functionals[functional]["dispersion"]) > 1
                and P["dispersion"] != "none"
            ):
                text += f" with the {P['dispersion']} dispersion correction."
            else:
                text += " with no dispersion correction."
        else:
            text = f"{calculation_type} using {method}."

        # Spin
        if P["spin-restricted"] == "yes":
            text += " The spin will be restricted to a pure eigenstate."
        elif P["spin-restricted"] == "default":
            if configuration is not None:
                if configuration.spin_multiplicity == 1:
                    text += " The spin will be restricted to a pure eigenstate."
                else:
                    text += " The spin will not be restricted and may not be a "
                    text += "proper eigenstate."
            else:
                text += " The spin will be restricted to a pure "
                text += "eigenstate for singlet states. Otherwise it will not "
                text += "be restricted and may not be a proper eigenstate."
        elif self.is_expr(P["spin-restricted"]):
            text += " Whether the spin will be restricted to a pure "
            text += "eigenstate will be determined by {P['spin-restricted']}"
        else:
            text += " The spin will not be restricted and may not be a "
            text += "proper eigenstate."

        # Plotting
        if isinstance(P["density"], str):
            density = P["density"] != "no"
        else:
            density = P["density"]

        if isinstance(P["orbitals"], str):
            orbitals = P["orbitals"] != "no"
        else:
            orbitals = P["orbitals"]

        if density:
            if orbitals:
                text += "\nThe alpha and beta electron, total, and spin densities, "
                text += f"and orbitals {P['selected orbitals']} will be plotted."
        elif orbitals:
            text += f"\nThe orbitals {P['selected orbitals']} will be plotted."

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self, calculation_type="energy", restart=None):
        """Get the input for an energy calculation for Psi4"""
        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        _, configuration = self.get_system_configuration(None)

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description = []
        self.description.append(
            __(
                self.description_text(PP, configuration=configuration),
                **PP,
                indent=self.indent,
            )
        )

        lines = []
        if calculation_type == "energy":
            lines.append("")
            lines.append("#" * 80)
            lines.append(f"# {self.header}")
            lines.append("#" * 80)

        # Figure out what we are doing!
        if P["level"] == "recommended":
            method_string = P["method"]
        else:
            method_string = P["advanced_method"]

        # Allow the full name, or the short name, or just pray.
        if method_string in psi4_step.methods:
            method = psi4_step.methods[method_string]["method"]
        else:
            method = method_string.lower()
            for key in psi4_step.methods:
                if psi4_step.methods[key]["method"] == method:
                    break
            else:
                method = method_string

        # lines.append('set scf_type df')
        # lines.append('set guess sad')

        multiplicity = configuration.spin_multiplicity
        spin_restricted = P["spin-restricted"]
        if spin_restricted == "default":
            if multiplicity == 1:
                restricted = True
            else:
                restricted = False
        elif spin_restricted == "yes":
            restricted = True
        else:
            restricted = False

        if method == "dft":
            if restricted:
                lines.append("set reference rks")
            else:
                lines.append("set reference uks")
        else:
            if restricted:
                if multiplicity == 1:
                    lines.append("set reference rhf")
                else:
                    lines.append("set reference rohf")
            else:
                lines.append("set reference uhf")

        if "freeze-cores" in P and P["freeze-cores"] == "yes":
            lines.append("set freeze_core True")
        else:
            lines.append("set freeze_core False")

        if P["stability analysis"]:
            lines.append("set stability_analysis True")

        # Convergence parameters and methods
        lines.append("set fail_on_maxiter False")  # Psi4 hangs! Cope in the plug-in

        lines.append(f"set maxiter {P['maximum iterations']}")

        if P["density convergence"] != "default":
            lines.append(f"set d_convergence {P['density convergence']}")
        if P["energy convergence"] != "default":
            lines.append(f"set e_convergence {P['energy convergence']}")

        if P["use damping"]:
            lines.append(f"set damping_percentage {P['damping percentage']}")
            lines.append(f"set damping_convergence {P['damping convergence']}")

        if P["use level shift"]:
            lines.append(f"set level_shift {P['level shift']}")
            lines.append(f"set level_shift_cutoff {P['level shift convergence']}")

        if P["use soscf"]:
            lines.append("set soscf True")
            lines.append(
                f"set soscf_start_convergence {P['soscf starting convergence']}"
            )
            lines.append(f"set soscf_conv {P['soscf convergence']}")
            lines.append(f"set soscf_max_iter {P['soscf max iterations']}")
            if P["soscf print iterations"]:
                lines.append("set soscf_print True")

        lines.append("")
        if method == "dft":
            if P["level"] == "recommended":
                functional_string = P["functional"]
            else:
                functional_string = P["advanced_functional"]

            # Allow the full name, or the short name, or just pray.
            if functional_string in psi4_step.dft_functionals:
                functional = psi4_step.dft_functionals[functional_string]["name"]
            else:
                functional = functional_string.lower()
                for key in psi4_step.dft_functionals:
                    if psi4_step.dft_functionals[key]["name"] == functional:
                        break
                else:
                    functional = functional_string

            if (
                P["dispersion"] != "none"
                and len(psi4_step.dft_functionals[functional_string]["dispersion"]) > 1
            ):
                functional = functional + "-" + P["dispersion"]
            if restart is None:
                lines.append(
                    f"Eelec, wfn = {calculation_type}('{functional}', return_wfn=True)"
                )
                lines.append(f"G = gradient('{functional}', ref_wfn=wfn)")
            else:
                if calculation_type == "gradient":
                    lines.append(
                        f"Eelec, wfn = energy('{functional}', return_wfn=True,"
                        f" restart_file='{restart}')"
                    )
                    lines.append(f"G = gradient('{functional}', ref_wfn=wfn)")
                else:
                    lines.append(
                        f"Eelec, wfn = {calculation_type}('{functional}', "
                        f"return_wfn=True, restart_file='{restart}')"
                    )
        else:
            if restart is None:
                lines.append(
                    f"Eelec, wfn = {calculation_type}('{method}', return_wfn=True)"
                )
                lines.append(f"G = gradient('{method}', ref_wfn=wfn)")
            else:
                if calculation_type == "gradient":
                    lines.append(
                        f"Eelec, wfn = energy('{method}', return_wfn=True, "
                        f" restart_file='{restart}')"
                    )
                    lines.append(f"G = gradient('{method}', ref_wfn=wfn)")
                else:
                    lines.append(
                        f"Eelec, wfn = {calculation_type}('{method}', return_wfn=True, "
                        f" restart_file='{restart}')"
                    )

        if calculation_type != "gradient":
            # Dump the properties to a json file
            filename = f"@{self._id[-1]}+properties.json"
            lines.append(
                f"""
try:
    oeprop(
        wfn,
        "MULTIPOLE(5)",
        "ESP_AT_NUCLEI",
        "LOWDIN_CHARGES",
        "MULLIKEN_CHARGES",
        "WIBERG_LOWDIN_INDICES",
        "MAYER_INDICES",
        "NO_OCCUPATIONS",
        title="PROP"
    )
except Exception:
    print("Failed to calculate properties")

variables = scalar_variables()
variables.update(wfn.scalar_variables())
arrays = array_variables()
for item in arrays:
    variables[item] = array_variable(item).np.tolist()
arrays = wfn.array_variables()
for item in arrays:
    variables[item] = wfn.array_variable(item).np.tolist()
variables["Eelec"] = Eelec
variables["energy"] = Eelec
try:
    variables["gradients"] = np.array(G).tolist()
except Exception as e:
    print("Problem with gradients!")
    print(e)

variables["_method"] = "{method}"
variables["_method_string"] = "{method_string}"


with open("{filename}", "w") as fd:
    json.dump(fix_multipoles(variables), fd, sort_keys=True, indent=3)
"""
            )

        # Orbital plots
        lines.append(self.plot_input())

        return "\n".join(lines)

    def analyze(self, indent="", data={}, out=[]):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access
        """
        # Read in the results from json
        directory = Path(self.directory)
        json_file = directory / "properties.json"
        if json_file.exists():
            with json_file.open() as fd:
                data = json.load(fd)

            # Put any requested results into variables or tables
            self.store_results(
                data=data,
                create_tables=self.parameters["create tables"].get(),
            )

            text = "The calculated energy is {Eelec:.6f} E_h."
        else:
            data = {}
            tmp = str(json_file)
            text = (
                "\nThere are no results from Psi4. Perhaps it "
                f"failed? Looking for {tmp}."
            )
            printer.normal(__(text, **data, indent=self.indent + 4 * " "))
            raise RuntimeError(text)

        # Write the structure locally for use in density and orbital plots
        system, configuration = self.get_system_configuration()
        obConversion = openbabel.OBConversion()
        obConversion.SetOutFormat("sdf")
        obMol = configuration.to_OBMol(properties="all")
        title = f"SEAMM={system.name}/{configuration.name}"
        obMol.SetTitle(title)
        sdf = obConversion.WriteString(obMol)
        path = directory / "structure.sdf"
        path.write_text(sdf)

        printer.normal(__(text, **data, indent=self.indent + 4 * " "))

    def plot_input(self):
        """Generate the input for plotting to cube files."""
        _, configuration = self.get_system_configuration(None)

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        tasks = []
        if P["density"]:
            tasks.append("density")

        orbitals = []
        if P["orbitals"]:
            tasks.append("orbitals")
            # and work out the orbitals
            txt = P["selected orbitals"]
            if txt == "all":
                pass
            else:
                # Which is the HOMO orbital?
                # This will not work with ECPs.
                n_electrons = (
                    sum(configuration.atoms.atomic_numbers) - configuration.charge
                )
                multiplicity = configuration.spin_multiplicity
                homo = (n_electrons - (multiplicity - 1)) // 2 + (multiplicity - 1)

                for chunk in txt.split(","):
                    chunk = chunk.strip()
                    if ":" in chunk or ".." in chunk:
                        if ":" in chunk:
                            first, last = chunk.split(":")
                        elif ".." in chunk:
                            first, last = chunk.split("..")
                        first = first.strip().upper()
                        last = last.strip().upper()

                        if first == "HOMO":
                            first = homo
                        elif first == "LUMO":
                            first = homo + 1
                        else:
                            first = int(first.removeprefix("HOMO").removeprefix("LUMO"))
                            if first < 0:
                                first = homo + first
                            else:
                                first = homo + 1 + first

                        if last == "HOMO":
                            last = homo
                        elif last == "LUMO":
                            last = homo + 1
                        else:
                            last = int(last.removeprefix("HOMO").removeprefix("LUMO"))
                            if last < 0:
                                last = homo + last
                            else:
                                last = homo + 1 + last

                        orbitals.extend(range(first, last + 1))
                    else:
                        first = chunk.strip().upper()

                        if first == "HOMO":
                            first = homo
                        elif first == "LUMO":
                            first = homo + 1
                        else:
                            first = int(first.removeprefix("HOMO").removeprefix("LUMO"))
                            if first < 0:
                                first = homo + first
                            else:
                                first = homo + 1 + first
                        orbitals.append(first)

        if len(tasks) == 0:
            return ""

        lines = []
        txt = "', '".join(tasks)
        lines.append("")
        lines.append("# Cube files for density and orbitals")
        lines.append(f"set cubeprop_tasks ['{txt}']")
        if len(orbitals) > 0:
            txt = ", ".join([f"{i}, {-i}" for i in orbitals])
            lines.append(f"set cubeprop_orbitals [{txt}]")

        lines.append("")
        lines.append("cubeprop(wfn)")
        lines.append(
            f"""
# Prefix the files with the substep number
paths = Path.cwd().glob('*.cube')
for path in paths:
    name = path.name
    newpath = path.with_name('@{self._id[-1]}+' + name)
    path.rename(newpath)
"""
        )

        return "\n".join(lines)



# -*- coding: utf-8 -*-

"""Setup and run Psi4"""

import logging
from pathlib import Path

import psi4_step
import seamm
import seamm.data
from seamm_util import units_class
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("psi4")


class Optimization(psi4_step.Energy):
    def __init__(
        self,
        flowchart=None,
        title="Optimization",
        extension=None,
        logger=logger,
    ):
        """Initialize the node"""

        logger.debug("Creating Optimization {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self._calculation = "optimization"
        self._model = None
        self._metadata = psi4_step.metadata
        self.parameters = psi4_step.OptimizationParameters()

        self.description = "A geometry optimization"

    def description_text(
        self,
        P=None,
        calculation_type="Geometry optimization",
        configuration=None,
    ):
        """Prepare information about what this node will do"""

        if not P:
            P = self.parameters.values_to_dict()

        text = super().description_text(
            P=P, calculation_type=calculation_type, configuration=configuration
        )

        added = "\nThe geometry optimization will use the {optimization method} "
        if P["max geometry steps"] == "default":
            added += "method, using the default maximum number of steps, which"
            added += " is based on the system size."
        else:
            added += "method, with no more than {max geometry steps} steps."

        if P["geometry convergence"] == "Custom":
            added += " The convergence criterion is"
        else:
            added += " The convergence criterion is '{geometry convergence}'."

        if P["recalc hessian"] != "never":
            added += " The Hessian will be recalculated every {recalc hessian}"
            added += " steps. Note that calculating the second derivatives is "
            added += "quite expensive!"

        return text + "\n" + __(added, **P, indent=4 * " ").__str__()

    def get_input(self, calculation_type="optimize"):
        """Get the input for an optimization calculation for Psi4"""
        _, configuration = self.get_system_configuration()

        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        # Check for atoms and bypass optimization
        if configuration.n_atoms == 1:
            return super().get_input()

        # references = self.parent.references

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description = []
        self.description.append(
            __(
                self.description_text(PP, configuration=configuration),
                **PP,
                indent=self.indent,
            )
        )

        lines = []
        lines.append("")
        lines.append("#" * 80)
        lines.append(f"# {self.header}")
        lines.append("#" * 80)
        lines.append("")
        lines.append("set opt_type min")
        if P["optimization method"] in psi4_step.optimization_methods:
            opt_method = psi4_step.optimization_methods[P["optimization method"]]
        else:
            opt_method = P["optimization method"]
        lines.append(f"set step_type {opt_method}")
        lines.append(f"set opt_coordinates {P['coordinates']}")
        max_steps = P["max geometry steps"]
        if max_steps == "default":
            max_steps = 6 * configuration.n_atoms
        if "nAtoms" in max_steps:
            n_atoms = configuration.n_atoms
            max_steps = max_steps.replace("nAtoms", str(n_atoms))
            max_steps = eval(max_steps)

        lines.append(f"set geom_maxiter {max_steps}")
        if P["geometry convergence"] == "Custom":
            pass
        else:
            if P["geometry convergence"] in psi4_step.optimization_convergence:
                convergence = psi4_step.optimization_convergence[
                    P["geometry convergence"]
                ]
            else:
                convergence = P["geometry convergence"]
            lines.append(f"set g_convergence {convergence}")

        if P["recalc hessian"] == "every step":
            lines.append("set full_hess_every 1")
        elif P["recalc hessian"] == "at beginning":
            lines.append("set full_hess_every 0")
        elif P["recalc hessian"] == "never":
            lines.append("set full_hess_every -1")
        else:
            lines.append(f"set full_hess_every {P['recalc hessian']}")
        lines.append(f"set hess_update {P['hessian update']}")
        lines.append("")

        # Add in the input from the energy part of things
        lines.append(super().get_input(calculation_type=calculation_type))

        return "\n".join(lines)



# -*- coding: utf-8 -*-
"""Global control parameters for Psi4
"""

import logging

from psi4_step import methods, dft_functionals
import seamm

logger = logging.getLogger(__name__)


class EnergyParameters(seamm.Parameters):
    """The control parameters for the energy."""

    parameters = {
        "level": {
            "default": "recommended",
            "kind": "string",
            "format_string": "s",
            "enumeration": ("recommended", "advanced"),
            "description": "The level of disclosure in the interface",
            "help_text": (
                "How much detail to show in the GUI. Currently 'recommended' "
                "or 'advanced', which shows everything."
            ),
        },
        "method": {
            "default": "Kohn-Sham (KS) density functional theory (DFT)",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": [x for x in methods if methods[x]["level"] == "normal"],
            "format_string": "s",
            "description": "Method:",
            "help_text": ("The computational method to use."),
        },
        "advanced_method": {
            "default": "Kohn-Sham (KS) density functional theory (DFT)",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": [x for x in methods],
            "format_string": "s",
            "description": "Method:",
            "help_text": ("The computational method to use."),
        },
        "functional": {
            "default": "B3LYP Hyb-GGA Exchange-Correlation Functional",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": [
                x for x in dft_functionals if dft_functionals[x]["level"] == "normal"
            ],
            "format_string": "s",
            "description": "DFT Functional:",
            "help_text": ("The exchange-correlation functional to use."),
        },
        "advanced_functional": {
            "default": "B3LYP Hyb-GGA Exchange-Correlation Functional",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": [x for x in dft_functionals],
            "format_string": "s",
            "description": "DFT Functional:",
            "help_text": ("The exchange-correlation functional to use."),
        },
        "dispersion": {
            "default": "d3bj",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": ["none", "d3bj", "d3mbj", "nl"],
            "format_string": "s",
            "description": "Dispersion correction:",
            "help_text": ("The dispersion correction to use."),
        },
        "spin-restricted": {
            "default": "default",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": ("default", "yes", "no"),
            "format_string": "s",
            "description": "Spin-restricted:",
            "help_text": (
                "Whether to restrict the spin (RHF, ROHF, RKS) or not "
                "(UHF, UKS)."
                " Default is restricted for singlets, unrestricted otherwise."
            ),
        },
        "use damping": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Damp the iterations:",
            "help_text": (
                "Whether to damp the iterations using a fraction of the previous "
                "density to damp oscillations."
            ),
        },
        "damping percentage": {
            "default": 20.0,
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "s",
            "description": "Percent damping:",
            "help_text": "Percent of previous density to use to damp oscillations.",
        },
        "damping convergence": {
            "default": 0.0,
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "s",
            "description": "Damp to convergence:",
            "help_text": "Convergence level to stop damping. 0 = always damp.",
        },
        "use level shift": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Use a level shift:",
            "help_text": ("Whether to use a level shift to help convergence."),
        },
        "level shift": {
            "default": 5.0,
            "kind": "float",
            "default_units": "E_h",
            "enumeration": tuple(),
            "format_string": "s",
            "description": "Level shift:",
            "help_text": "The amount to shift the occupied orbitals down.",
        },
        "level shift convergence": {
            "default": 0.01,
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "s",
            "description": "Stop when converged to:",
            "help_text": "Convergence level to stop level shifting.",
        },
        "use soscf": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Use second order SCF:",
            "help_text": ("Whether to use the second order SCF to help convergence."),
        },
        "soscf starting convergence": {
            "default": 0.01,
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "s",
            "description": "Start when converged to:",
            "help_text": "Convergence level to start second order SCF.",
        },
        "soscf convergence": {
            "default": 0.001,
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "s",
            "description": "Microiteration convergence:",
            "help_text": "Convergence level for SOSCF microiterations.",
        },
        "soscf max iterations": {
            "default": 5,
            "kind": "integer",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "s",
            "description": "Maximum microiterations:",
            "help_text": "Maximum number of SOSCF microiterations.",
        },
        "soscf print iterations": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Print microiterations:",
            "help_text": "Print information about the SOSCF microiterations.",
        },
        "density convergence": {
            "default": "default",
            "kind": "float",
            "default_units": "",
            "enumeration": ("default",),
            "format_string": "s",
            "description": "Density convergence criterion:",
            "help_text": (
                "Criterion for convergence of the density, default 10^-6 for "
                "SCF, 10^-8 optimization."
            ),
        },
        "energy convergence": {
            "default": "default",
            "kind": "float",
            "default_units": "",
            "enumeration": ("default",),
            "format_string": "s",
            "description": "Energy convergence criterion:",
            "help_text": (
                "Criterion for convergence of the energy, default 10^-6 for "
                "SCF, 10^-8 optimization."
            ),
        },
        "convergence error": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Error if not converged:",
            "help_text": "Whether to throw an error if not converged.",
        },
        "maximum iterations": {
            "default": 100,
            "kind": "integer",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "s",
            "description": "Maximum iterations:",
            "help_text": "Maximum number of SCF iterations.",
        },
        "freeze-cores": {
            "default": "yes",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Freeze core orbitals:",
            "help_text": (
                "Whether to freeze the core orbitals in correlated " "methods"
            ),
        },
        "stability analysis": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "s",
            "description": "Do stability analysis:",
            "help_text": ("Analyze the stability of the SCF/DFT wavefunction."),
        },
        "results": {
            "default": {},
            "kind": "dictionary",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "",
            "description": "results",
            "help_text": ("The results to save to variables or in " "tables. "),
        },
        "create tables": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Create tables as needed:",
            "help_text": (
                "Whether to create tables as needed for "
                "results being saved into tables."
            ),
        },
    }

    output = {
        "density": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Plot total density:",
            "help_text": "Whether to plot the total charge density.",
        },
        "orbitals": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Plot orbitals:",
            "help_text": "Whether to plot orbitals.",
        },
        "selected orbitals": {
            "default": "-1, HOMO, LUMO, +1",
            "kind": "string",
            "default_units": "",
            "enumeration": ("all", "-1, HOMO, LUMO, +1"),
            "format_string": "",
            "description": "Selected orbitals:",
            "help_text": "Which orbitals to plot.",
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={
                **EnergyParameters.parameters,
                **EnergyParameters.output,
                **defaults,
            },
            data=data,
        )



# -*- coding: utf-8 -*-
"""Global control parameters for Psi4
"""

import logging
import seamm

logger = logging.getLogger(__name__)


class InitializationParameters(seamm.Parameters):
    """The control parameters for initializing Psi4"""

    parameters = {
        "basis": {
            "default": "6-31G**",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "6-31G",
                "6-31G(d)",
                "6-31G(d,p)",
                "6-31+G",
                "6-31+G(d)",
                "6-31+G(d,p)",
                "6-311G",
                "6-311G(d)",
                "6-311G(d,p)",
                "6-311+G",
                "6-311+G(d)",
                "6-311+G(d,p)",
                "cc-pVDZ",
                "cc-pVTZ",
                "cc-pVQZ",
                "def2-SV(P)",
                "def2-SVP",
                "def2-TZVP",
                "def2-TZVPP",
                "def2-QZVP",
                "def2-QZVPP",
            ),
            "format_string": "s",
            "description": "Basis:",
            "help_text": ("The basis set to use."),
        },
        "symmetry_tolerance": {
            "default": "0.00001",
            "kind": "float",
            "default_units": None,
            "enumeration": tuple(),
            "format_string": ".1e",
            "description": "Symmetry tolerance:",
            "help_text": "The tolerance used when determining the symmetry.",
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={**InitializationParameters.parameters, **defaults}, data=data
        )



# -*- coding: utf-8 -*-

"""Non-graphical part of the Thermochemistry step in a Psi4 flowchart
"""

import json
import logging
from pathlib import Path
import pkg_resources

from tabulate import tabulate

import psi4_step
import molsystem
import seamm
from seamm_util import Q_, units_class
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: "job" and "printer". "job" send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# "printer" sends output to the file "step.out" in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("Psi4")

# Add this module's properties to the standard properties
path = Path(pkg_resources.resource_filename(__name__, "data/"))
csv_file = path / "properties.csv"
if path.exists():
    molsystem.add_properties_from_file(csv_file)


class Thermochemistry(psi4_step.Energy):
    """
    The non-graphical part of a Thermochemistry step in a flowchart.

    Attributes
    ----------
    parser : configargparse.ArgParser
        The parser object.

    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : ThermochemistryParameters
        The control parameters for Thermochemistry.

    See Also
    --------
    TkThermochemistry,
    Thermochemistry, ThermochemistryParameters
    """

    def __init__(
        self,
        flowchart=None,
        title="Thermochemistry",
        extension=None,
        logger=logger,
    ):
        """A substep for Thermochemistry in a subflowchart for Psi4.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
        flowchart: seamm.Flowchart
            The non-graphical flowchart that contains this step.

        title: str
            The name displayed in the flowchart.
        extension: None
            Not yet implemented
        logger : Logger = logger
            The logger to use and pass to parent classes

        Returns
        -------
        None
        """
        logger.debug(f"Creating Thermochemistry {self}")

        super().__init__(
            flowchart=flowchart,
            title=title,
            extension=extension,
            logger=logger,
        )

        self._calculation = "thermochemistry"
        self._model = None
        self._metadata = psi4_step.metadata
        self.parameters = psi4_step.ThermochemistryParameters()

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return psi4_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return psi4_step.__git_revision__

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
        P: dict
            An optional dictionary of the current values of the control
            parameters.
        Returns
        -------
        str
            A description of the current step.
        """
        if not P:
            P = self.parameters.values_to_dict()

        text = super().description_text(P=P, calculation_type="Thermochemistry")

        added = (
            "\nThe thermodynamic functions will be calculated at temperature {T} and "
            "pressure {P}."
        )

        return text + "\n" + __(added, **P, indent=4 * " ").__str__()

    def get_input(self, calculation_type="frequency"):
        """Get the input for an optimization calculation for Psi4"""
        _, configuration = self.get_system_configuration()

        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        # references = self.parent.references

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description = []
        self.description.append(__(self.description_text(PP), **PP, indent=self.indent))
        # Figure out what we are doing! The method is HF, B3LYP, CCSD, etc.
        if P["level"] == "recommended":
            method_string = P["method"]
        else:
            method_string = P["advanced_method"]

        if method_string in psi4_step.methods:
            method = psi4_step.methods[method_string]["method"]
        else:
            method = method_string

        if method == "dft":
            if P["level"] == "recommended":
                functional_string = P["functional"]
            else:
                functional_string = P["advanced_functional"]
            method = psi4_step.dft_functionals[functional_string]["name"]
            if (
                P["dispersion"] != "none"
                and len(psi4_step.dft_functionals[functional_string]["dispersion"]) > 1
            ):
                method = method + "-" + P["dispersion"]

        lines = []
        lines.append("")
        lines.append("#" * 80)
        lines.append(f"# {self.header}")
        lines.append("#" * 80)
        lines.append("")
        # lines.append("initial.find_point_group(tolerance=1.0e-5)")
        lines.append("initial.symmetrize(1.0e-5)")
        lines.append("point_group = initial.point_group().symbol()")
        lines.append("")

        if not P["use existing parameters"]:
            # Add in the input from the energy part of things
            lines.append(super().get_input(calculation_type=calculation_type))
        else:
            lines.append(f"Eelec, wfn = frequency('{method}', return_wfn=True)")

            # Orbital plots
            lines.append(self.plot_input())

        lines.append(
            f"""
set writer_file_label thermo
set hessian_write on

thermo_vibinfo = vibanal_wfn(wfn)
path = Path(core.get_writer_file_prefix('thermo') + '.vibrec')
with path.open() as fd:
    tmp = json.load(fd)
tmp2 = dict()
for key, value in tmp.items():
    if type(value) == str:
        tmp2[key] = json.loads(value)
    else:
        tmp2[key] = value

with path.with_name('@{self._id[-1]}+thermochemistry.json').open('w') as fd:
    json.dump(tmp2, fd, sort_keys=True, indent=3)
"""
        )

        return "\n".join(lines)

    def analyze(self, indent="", **kwargs):
        """Do any analysis of the output from this step.

        Also print important results to the local step.out file using
        "printer".

        Parameters
        ----------
        indent: str
            An extra indentation for the output
        """
        # Read in the results from json
        directory = Path(self.directory)
        json_file = directory / "thermochemistry.json"
        if json_file.exists():
            with json_file.open() as fd:
                tmp = json.load(fd)

            # Process the data, changing units
            d = {}
            metadata = self.metadata["results"]
            for key, value in tmp.items():
                if key in metadata:
                    from_units = value["units"].replace("Eh", "E_h")
                    to_units = metadata[key]["units"]
                    if from_units == to_units:
                        d[key] = value["data"]
                    else:
                        d[key] = Q_(value["data"], from_units).m_as(to_units)

            # Put any requested results into variables or tables
            self.store_results(data=d, create_tables=True)

            # And the output
            table = {
                "": [
                    "Units",
                    "Electronic",
                    "Translational",
                    "Rotational",
                    "Vibrational",
                    "Total Correction",
                    "Total (E_h)",
                ],
                "S": [
                    "J/mol/K",
                    round(d["S_elec"], 2),
                    round(d["S_trans"], 2),
                    round(d["S_rot"], 2),
                    round(d["S_vib"], 2),
                    round(d["S_tot"], 2),
                ],
                "Cv": [
                    "J/mol/K",
                    round(d["Cv_elec"], 2),
                    round(d["Cv_trans"], 2),
                    round(d["Cv_rot"], 2),
                    round(d["Cv_vib"], 2),
                    round(d["Cv_tot"], 2),
                ],
                "Cp": [
                    "J/mol/K",
                    round(d["Cp_elec"], 2),
                    round(d["Cp_trans"], 2),
                    round(d["Cp_rot"], 2),
                    round(d["Cp_vib"], 2),
                    round(d["Cp_tot"], 2),
                ],
                "E": [
                    "kJ/mol",
                    round(d["E_elec"], 2),
                    round(d["E_trans"], 2),
                    round(d["E_rot"], 2),
                    round(d["E_vib"], 2),
                    round(d["E_corr"], 2),
                    round(d["E_tot"], 6),
                ],
                "H": [
                    "kJ/mol",
                    round(d["H_elec"], 2),
                    round(d["H_trans"], 2),
                    round(d["H_rot"], 2),
                    round(d["H_vib"], 2),
                    round(d["H_corr"], 2),
                    round(d["H_tot"], 6),
                ],
                "G": [
                    "kJ/mol",
                    round(d["G_elec"], 2),
                    round(d["G_trans"], 2),
                    round(d["G_rot"], 2),
                    round(d["G_vib"], 2),
                    round(d["G_corr"], 2),
                    round(d["G_tot"], 6),
                ],
            }

            text = ""
            tmp = tabulate(
                table,
                headers="keys",
                tablefmt="simple_outline",
                disable_numparse=True,
                numalign="decimal",
                stralign="center",
                colalign=(
                    "center",
                    "decimal",
                    "decimal",
                    "decimal",
                    "decimal",
                    "decimal",
                    "decimal",
                ),
            )
            length = len(tmp.splitlines()[0])
            text += "\n"
            parameters = self.parameters.current_values_to_dict(
                context=seamm.flowchart_variables._data
            )
            T = parameters["T"]
            P = parameters["P"]
            text += f"Thermodynamic Functions at {T:.2f~P} and {P:.2f~P}".center(length)
            text += "\n"
            text += tmp
            text += "\n"
            printer.normal(__(text, indent=8 * " ", wrap=False, dedent=False))
        else:
            text = (
                "\nThere are no thermochemistry results from Psi4. Perhaps it "
                f"failed? Looking for {str(json_file)}."
            )
            printer.normal(__(text, indent=self.indent + 4 * " "))
            raise RuntimeError(text)



# -*- coding: utf-8 -*-

"""
psi4_step
A step for Psi4 in a SEAMM flowchart
"""

# Bring up the classes so that they appear to be directly in
# the psi4_step package.

from .psi4_metadata import methods  # noqa: F401
from .psi4_metadata import dft_functionals  # noqa: F401
from .psi4_metadata import optimization_methods  # noqa: F401
from .psi4_metadata import optimization_convergence  # noqa: F401
from .psi4_metadata import metadata  # noqa: F401

from .psi4 import Psi4  # noqa: F401
from .psi4_parameters import Psi4Parameters  # noqa: F401
from .psi4_step import Psi4Step  # noqa: F401
from .tk_psi4 import TkPsi4  # noqa: F401

from .initialization import Initialization  # noqa: F401
from .initialization_parameters import InitializationParameters  # noqa: F401
from .initialization_step import InitializationStep  # noqa: F401
from .tk_initialization import TkInitialization  # noqa: F401

from .energy import Energy  # noqa: F401
from .energy_parameters import EnergyParameters  # noqa: F401
from .energy_step import EnergyStep  # noqa: F401
from .tk_energy import TkEnergy  # noqa: F401

from .optimization import Optimization  # noqa: F401
from .optimization_parameters import OptimizationParameters  # noqa: F401
from .optimization_step import OptimizationStep  # noqa: F401
from .tk_optimization import TkOptimization  # noqa: F401

# from .accelerated_optimization import AcceleratedOptimization  # noqa: F401
# from .accelerated_optimization_parameters import (  # noqa: F401
#     AcceleratedOptimizationParameters,
# )
# from .accelerated_optimization_step import (  # noqa: F401
#     AcceleratedOptimizationStep,
# )
# from .tk_accelerated_optimization import (  # noqa: F401
#     TkAcceleratedOptimization,
# )

from .thermochemistry_step import ThermochemistryStep  # noqa: F401
from .thermochemistry import Thermochemistry  # noqa: F401
from .thermochemistry_parameters import ThermochemistryParameters  # noqa: F401
from .tk_thermochemistry import TkThermochemistry  # noqa: F401

# Handle versioneer
from ._version import get_versions

__author__ = """Paul Saxe"""
__email__ = "psaxe@molssi.org"
versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions
